#!/usr/bin/env python3
import json
import yaml
import os
import re
import sys
import subprocess
import tempfile

HEADER = r"""# Naevpedia
# Copyright (C) YEAR Naev Dev Team
# This file is distributed under the same licence as the naev package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: https://github.com/naev/naev/issues\n"
"POT-Creation-Date: YEAR-MO-DA HO:MI+ZONE\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@li.org>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"

"""

def needs_translation( line ):
   if not line:
      return False
   ls = line.strip()
   if ls[:2]=="<%" and ls[-2:]=="%>":
      return False
   return True

from tempfile import NamedTemporaryFile
with open(sys.argv[1],"w") as fout:
   def print_line( fn, i, line ):
      quoted_escaped_line = json.dumps(line, ensure_ascii=False)
      fout.write('#: {}:{}\nmsgid {}\nmsgstr ""\n\n'.format(fn, i+1, quoted_escaped_line))

   # Want reproducible stuff
   filenames = sys.argv[2:].copy()
   filenames.sort()

   fout.write( HEADER )
   for fn in filenames:
      with open( fn, 'r' ) as f:
         fn = re.sub('.*/dat', 'dat', fn)
         d = f.read()
         # Remove meta-data header, TODO parse
         m = d.split('---\n')
         n = 0
         if len(m) > 2:
            y = yaml.safe_load( m[1] )
            n = 2 + len(m[1].splitlines())
            if 'title' in y:
               print_line( fn, 1, y['title'] )
            d = ''.join(m[2:])
         # Preprocess to get rid of Lua blocks
         luas = re.findall( "<%[^=].*?%>", d, flags=re.S )
         tmp = NamedTemporaryFile(mode='wt', delete=False)
         tmpnam = tmp.name
         for l in luas:
            tmp.write(l[2:-2]+"\n")
         tmp.close()

         args = [ "xgettext", tmpnam, "-o", "-", "-L", "Lua", "--omit-header" ]
         ret = subprocess.run( args, capture_output=True, text=True )
         os.remove(tmpnam)
         res = [
            lin.replace(tmp.name+':', 'stdin:') if lin[:2] =='#:' else lin
            for lin in ret.stdout.split('\n')
         ]
         fout.write( '\n'.join(res) + "\n" )
         # Replace Lua blocks with <%lua%>
         d = re.sub( "<%[^=].*?%>", "<%lua%>", d, flags=re.S )
         # Go over lines
         l = d.splitlines()
         for i, line in enumerate(l):
            if needs_translation( line ):
               print_line( fn, n+i, line )



#!/usr/bin/env python3


from math import cos, sin, pi, sqrt, asin

EPS = 0.0000001


# vec

class _vec(tuple):
   def __add__( self, other ):
      return _vec([x+y for (x, y) in zip(self, other)])

   def __sub__( self, other ):
      return self + other*-1.0

   def orth( self ):
      return _vec((-self[1], self[0]))

   def _rotate( self, sa, ca ):
      return _vec((self[0]*ca - self[1]*sa, self[0]*sa + self[1]*ca))

   def __mul__( self, other ):
      if isinstance(other, _vec):
         # Dot product
         return sum([a*b for (a,b) in zip(self, other)])
      elif isinstance(other, _transf):
         # Apply transformation
         return other(self)
      else:
         # External product
         return _vec([x*other for x in self])

   __rmul__ = __mul__
   __hash__ = tuple.__hash__

   def __neg__( self ):
      return self * -1.0

   def __truediv__( self, other ):
      if isinstance(other, _vec):
         # The transformation that turns other into self.
         return _transf(other, self)
      else:
         return self * (1.0/other)

   def __round__( self, dig = 0 ):
      return _vec([round(x, dig) for x in self])

   def __str__( self ):
      return str(tuple([int(a) if int(a) == a else a for a in self]))

   def __repr__( self ):
      return 'vec' + tuple.__repr__(self)

   def __eq__( self, other ):
      if isinstance(other, tuple):
         other = _vec(other)
      if not isinstance(other, _vec):
         return False
      else:
         d = self - other
         return d*d < EPS*EPS

   def rotate_rad( self, angle ):
      return self._rotate(sin(angle), cos(angle))

   def rotate( self, degrees ):
      return self.rotate_rad(degrees / 180.0 * pi)

   def size( self ):
      return sqrt(self * self)

   def normalize( self, new_size = 1.0 ):
      if (size:= self.size()) < EPS:
         return self
      else:
         return self / size * new_size

class _transf:
   """
   A rotation and a scaling.
   Defined by a pair of vectors (before trans, after trans)
   """
   def __init__( self, v1, v2 ):
      l1 = v1.size()
      self.fact = v2.size()/l1
      v1, v2 = v1.normalize(), v2.normalize()
      self.vec = v1[0]*v2[1] - v1[1]*v2[0]
      if self.vec > 1.0:
         self.vec = 1.0
      elif self.vec < -1.0:
         self.vec = -1.0
      if v1*v2 < 0:
         self.vec  = -self.vec

   def __str__( self ):
      return str({'fact': self.fact, 'vec': self.vec})

   def __call__( self, other ):
      sa = self.vec
      return other._rotate(sa, sqrt(1.0 - sa*sa)) * self.fact


def vec( *args ):
   if len(args) == 0:
      args = (0, 0)
   elif len(args) == 1:
      args = args[0]
   return _vec((float(x) for x in args))

def transf():
   return vec(1, 0) / vec(1, 0)


# general geometry

class line:
   def __init__( self, P, v ):
      self.P = P
      self.v = v

   def __sub__( self, P ):
      u = self.P - P
      v = self.v
      u-= (u * v) * v / (v * v)
      return u

   def closest(self, P):
      return P + (self - P)

   # use & to call
   def __and__( self, other ):
      if isinstance(other, line):
         u = self.v
         v = other.v.orth()
         s = u*v
         if abs(s) < EPS:
            return None
         else:
            return self.P + (other.P - self.P) * v / s * u
      elif isinstance(other, segment):
         P = self & other.line()
         if P and P in other:
            return P
         else:
            return None
      elif isinstance(other, circle):
         u = self - other.centre
         if (cl := (other.centre + u)) not in other:
            return []
         else:
            d = sqrt(other.radius*other.radius - u*u)
            if d < EPS:
               return [cl]
            else:
               # both might be close to null vector
               if self.v.size() > u.size():
                  v = self.v
               else:
                  v = u.orth()
               v = v.normalize(d)
               return [cl - v, cl + v]
      else:
         return NotImplemented

   def __repr__( self ):
      return 'line' + repr((self.P, self.v))

   __str__ = __repr__

class segment:
   def __init__( self, A, B ):
      self.A, self.B = A, B

   def line( self ):
      return line(self.B, self.B - self.A)

   def bisector( self ):
      return line((self.A + self.B) / 2.0, (self.B - self.A).orth())

   def __and__ ( self, other ):
      if isinstance(other, line):
         return other & self
      elif isinstance(other, segment):
         if self.line() & other:
            return other.line() & self
         else:
            return None
      else:
         return NotImplemented
   def points( self ):
      return set([self.A, self.B])

   # !!! only test whether C strictly between
   # the lines orth to (A B) though resp A and B.
   # If you also want A, B, C to be colinear, you have to do the test separately.
   def __contains__( self, C ):
      u = self.B - self.A
      return ((C-self.A)*u) * ((C-self.B)*u) < - EPS

   def closest( self, P ):
      H = self.line().closest(P);
      if H in self:
         return H
      elif self.A in segment(self.B, H):
         return self.A
      else:
         return self.B

class circle:
   def __init__( self, centre = vec(), radius = 0.0, strict = False ):
      self.centre = centre
      self.radius = radius
      self.strict = False

   def __repr__( self ):
      return 'circle' + repr((self.centre, self.radius))

   __str__ = __repr__

   def strictness( self, strict ):
      prv, self.strict = self.strict, strict
      return prv

   def __eq__( self, other ):
      if not isinstance(other, circle):
         return False
      d = self.radius - other.radius
      return self.centre == other.centre and d*d < EPS*EPS

   def __contains__( self, P ):
      u = P - self.centre
      epsilon = -EPS if self.strict else +EPS
      return u * u <= self.radius * self.radius * (1.0 + epsilon)

def inscribed( A, B, C ):
   la = line(A , (B - A).normalize() + (C - A).normalize())
   lb = line(B , (C - B).normalize() + (A - B).normalize())
   P = la & lb
   return circle(P, (line(A, B - A) - P).size())

def inscribed_lines( l1, l2, l3 ):
   P3, P1, P2 = (l1 & l2, l2 & l3, l3 & l1)
   if P1 and P2 and P3:
      return inscribed(P3, P1, P2)
   # TODO: manage other cases

def circumscribed( A, B, C ):
   la = line ((B + C) / 2.0, (B - C).orth())
   lb = line ((C + A) / 2.0, (C - A).orth())
   P = la & lb
   return circle(P, (A - P).size())

# subsets of S that have size n
def subsets( S, n ):
   if len(S) == n or n == 0:
      yield tuple(S[:n])
   else:
      S0, S = S[0], S[1:]
      for i in subsets(S, n):
         yield i
      for i in subsets(S, n-1):
         yield (S0,) + i

def bounding_circle( L ):
   def candidates( ):
      for (i, j) in subsets(L, 2):
         yield circle((i+j)/2, (i-j).size()/2)
      for (i, j, k) in subsets(L, 3):
         yield circumscribed(i, j, k)
   best = None
   for c in candidates():
      if (best is None or best.radius > c.radius):
         best, prv = c, best
         for p in L:
            if p not in c:
               best = prv
               break
   return best

def acute_T( A, B, C ):
   return (B-A)*(B-C)>=0.0-EPS and (C-B)*(C-A)>=0.0-EPS and (A-C)*(A-B)>=0.0-EPS

def interstices( edges, vertices = [], povfp = None ):
   S = [segment(*t) for t in edges]
   sides = set([(b,a) for (a,b) in edges] + edges)
   L = list(set([a for (a, b) in edges] + [b for (a, b) in edges] + vertices))

   def candidates( ):
      # line, line, line -> inscribed
      for t in subsets(S, 3):
         lines = tuple(x.line() for x in t)
         if C := inscribed_lines(*lines):
            c = C.centre
            if c in t[0] and c in t[1] and c in t[2]:
               yield C, tuple(l.closest(c) for l in lines), 0
      # line, line, point -> manual solution
      # TODO: manage the case the lines are parallel
      for t1, t2 in subsets(S, 2):
         l1, l2 = t1.line(), t2.line()
         I = l1 & l2
         for P in set(L).difference(t1.points(), t2.points()):
            u = l1.v * (P-I) * l1.v
            v = l2.v * (P-I) * l2.v
            if u*u < EPS or v*v < EPS:
               continue
            O = I + u.normalize() + v.normalize()
            UC = circle(O, (l1 - O).size())
            li = segment(I, P).line()
            if (P2 := (li & UC)) == []:
               continue
            P2 = P2[-1] if (P2[0]-I).size()<=(P2[-1]-I).size() else P2[0]
            ratio = ((P-I) * (P-I)) / ((P2-I) * (P-I))
            C = circle(I + (O-I)*ratio, UC.radius * abs(ratio))
            c = C.centre
            if c in t1 and c in t2:
               yield C, (P, l1.closest(c), l2.closest(c)), 1
      # point, point, line -> the point point line problem of Apollonius (PPL)
      for t in S:
         l1 = t.line()
         for (A, B) in subsets(list(set(L).difference(t.points())), 2):
            if (A, B) in sides:
               continue
            AB = segment(A, B)
            l2 = AB.line()
            O = l1 & l2
            if O is None:
               T2 = AB.bisector() & l1
            else:
               C = (A + (O + (O - B))) / 2.0
               C1 = circle(C, (C-A).size())
               T = line(O, l2.v.orth()) & C1
               if T == []:
                  continue
               T = T[-1] # take one intersection
               v = l1.v.normalize((T-O).size())
               if v*((A+B)/2-O) <= 0.0:
                  v = -v
               T2 = O + v
            if T2 in t:
               yield circumscribed(A, B, T2), (A, B, T2), 2
      # point, point, point -> circumscribed
      for t in subsets(L, 3):
         if sides.isdisjoint(subsets(t,2)):
            yield circumscribed(*t), t, 3
   out = []
   for c, T, n in candidates():
      if acute_T(*T):
         out.append(c)
         prev = c.strictness(True)
         for t in S:
            if t.closest(c.centre) in c:
               if povfp and n!=3:
                  print_circle(povfp, out.pop(), "Magenta")
               else:
                  out.pop()
               break
         c.strictness(prev)
   prv = None
   for C in sorted(out, key = lambda c: c.radius, reverse = True):
      if C != prv:
         prv = C
         yield C

def bounded_circles( L, povfp = None ):
   sides = list(zip(L, L[1:]+L[:1]))
   ext = vec(min([x for (x,_) in L]), min([y for (_,y) in L])) - vec(1, 1)
   for C in interstices(sides , povfp = povfp):
      l = segment(C.centre, ext)
      if len([1 for s in sides if l & segment(*s)]) % 2 == 1:
         yield C

# bounding boxes

class bb:
   def __init__( self ):
      self.empty = True

   def __contains__( self, v):
      return not (
         self.empty or
         self.minx > v[0] or self.maxx < v[0] or
         self.miny > v[1] or self.maxy < v[1]
      )

   def __iadd__( self, t ):
      if self.empty:
         (self.maxx, self.maxy) = t
         (self.minx, self.miny) = t
         self.empty = False
      else:
         if self.minx > t[0]:
            self.minx = t[0]
         elif self.maxx < t[0]:
            self.maxx = t[0]
         if self.miny > t[1]:
            self.miny = t[1]
         elif self.maxy < t[1]:
            self.maxy = t[1]
      return self

   def size( self ):
      return vec(self.maxx - self.minx, self.maxy - self.miny)

   def __imul__( self, other ):
      other /= 2.0
      xd, yd = self.size()
      self.minx -= xd * other
      self.maxx += xd * other
      self.miny -= yd * other
      self.maxy += yd * other
      return self

   def mini( self ):
      return vec(self.minx, self.miny) if not self.empty else None

   def maxi( self ):
      return vec(self.maxx, self.maxy) if not self.empty else None

   def __str__( self ):
      return str(round(self.mini())) + ':' + str(round(self.maxi()))

def symmetry( fst, snd = None ):
   if isinstance(fst, _vec) and isinstance(snd, _vec):
      fst = segment(fst, snd)
   if isinstance(fst, segment):
      fst = fst.line()
   if not isinstance(fst, line) and not isinstance(fst, _vec):
      raise ValueError("symmetry accepts 1 or 2 points, 1 segment or 1 line.")
   return lambda p: p + 2.0 * (fst - p)


# example

if __name__ == '__main__':
   from sys import argv, exit
   if '-h' in argv[1:] or '--help' in argv[1:]:
      print(
         'usage:  '+ argv[0]+'\n'
         '  Use povray to generate a test figure.'
      )
      exit(0)

   from tempfile import NamedTemporaryFile
   from subprocess import run
   from random import random
   import os
   pov_head = """
   #version 3.7;
   global_settings{ assumed_gamma 1.6 ambient_light 1.0 }
   #include "colors.inc"
   camera{
      orthographic
      sky         <0,1,0>
      location    z
      up          -2*y
      right       2*x*image_width/image_height
      look_at     0
   }
   light_source{ <10,10,100> White }
   //cylinder{ <0,0,0>, <0,1.0,0>, 0.002 pigment{color Green} }
   //cylinder{ <0,0,0>, <1.0,0,0>, 0.002 pigment{color Red} }
   """
   getpath = lambda *x: os.path.realpath(os.path.join(*x))
   script_dir = os.path.dirname(__file__)
   povfp = NamedTemporaryFile(mode='wt', buffering=-1, suffix='.pov')
   povname = povfp.name
   povfp.write(pov_head)

   point_it = lambda P: '<'+str(P[0])+','+str(P[1])+',0>'
   def print_point( fp, P, c = 'White' ):
      fp.write('sphere{ ' + point_it(P) + ', 0.002 pigment{color ' + c + '}}\n')

   def print_seg( fp, A, B, c = 'White' ):
      fp.write('cylinder{ ' + point_it(A) + ','+ point_it(B))
      fp.write( ', 0.002 pigment{color ' + c + '}}\n')

   def print_cross( fp, A, c = 'White', alt_cross = False ):
      u = (0.016 * vec(1,0)) if alt_cross else (0.008 * vec(1,1))
      print_seg(fp, A - u, A + u, c)
      print_seg(fp, A - u.orth(), A + u.orth(), c)

   def print_poly( fp, L, c = 'White' ):
      [print_seg(fp, *t, c) for t in zip(L,L[1:]+L[:1])]
      [print_point(povfp, P) for P in L]

   def print_circle( fp, C, c = 'White', alt_cross = False ):
      print_cross(fp, C.centre, c, alt_cross)
      fp.write('torus{ ' + str(C.radius) + ', 0.002 ')
      fp.write('rotate 90*x ' + 'translate '+point_it(C.centre))
      fp.write('pigment{color ' + c + '}}\n')

   ranv = lambda: vec(random()-0.5, random()-0.5)

   W, H = 3, 2
   Pols = [[[ranv() for _ in range(4)] for _ in range(W)] for _ in range(H)]
   Pols[0][0] = [0.35*vec(*t) for t in [
      (-0.96,-1), (0.91,-1), (1.02,1),
      (0.2,-0.52), (0.35,-0.54), (0.01, 1.0), (-0.33,-0.51), (-0.43,-0.5),
      (-1.03,1.03)
   ]]
   u = vec(0,1)
   v = vec(0,1.01).rotate(120)
   w = vec(0,1.02).rotate(-120)
   Pols[0][1] = [0.35*t for t in [u,-0.3*v,w,-0.3*u,v,-0.3*w]]
   Pols[0][2] = [vec(0.35,0).rotate(360.0/5.0*i) for i in range(5)]
   # TODO: manage parallel cases
   #u, v = vec(0.3,-0.2), vec(0.3,0.2)
   #Pols[0][2] = [u, v, -u, -v]

   for j in range(W):
      u = 16.0/9.0*(1.0*j/(W-1)-0.5)
      for i in range(H):
         v = 1.0 * i / (H-1) - 0.5
         Pol = [P - vec(u, v) for P in Pols[i][j]]
         B = bounding_circle(Pol)
         if B:
            d = B.centre - vec(u, v)
            Pol = [P - d for P in Pol]
            B.centre -= d
            print_circle(povfp, B, 'Blue', alt_cross = True)
         print_poly(povfp, Pol, 'Yellow')
         colour = 'Orange'
         for B in bounded_circles(Pol, None):
            print_circle(povfp, B, colour)
            colour = 'Red'

   povfp.flush()
   run(['povray', '+W1280', '+H720', '+A0.1', '+AM2', '+R3', '+J',
      '+I' + povfp.name, '+O' + getpath(script_dir, 'geometry.png') ])
   povfp.close()



#!/usr/bin/env python3
import helper as h
N_ = lambda text: text
data = h.read()

data['outfit']['@name'] = N_('Reaver Cannon')

general = data['outfit']['general']
general['rarity'] = '2'
general['description'] = N_('The complex engineering specifications and difficulty of mass production doomed the Reaver Cannon from the start. However, in the right hands, given the extreme long range and high penetration for a weapon of its size, it can be very deadly.')
general['$price'] += 130e3
general['$mass'] += 2
general['$cpu'] = round(general['$cpu'] * 1.1)

specific = data['outfit']['specific']
specific['$range'] *= 1.5
specific['$falloff'] *= 1.2
specific['$delay'] *= 2
specific['$energy'] *= 2.0
specific['$swivel'] *= 1.5
specific['$trackmax'] = round(specific['$trackmax'] * 0.8333333333)
specific['lua'] = 'outfits/lib/matrix_sell.lua'
specific['gfx'] = {
   '@type' : 'shader',
   '@size' : '10',
   '@col_size' : '8',
   '#text' : 'particles/reaver.frag',
}

damage = specific['damage']
damage['$physical'] *= 2
damage['$penetrate'] = round(damage['$penetrate'] * 1.3)

data.save()



#!/usr/bin/env python3

from os import path
import sys
import math

N_ = lambda text: text


script_dir = path.join(path.dirname(__file__), '..', '..', '..', 'utils', 'outfits')
sys.path.append(path.realpath(script_dir))
from outfit import outfit

def get_outfit_dict( nam, core= False, doubled= False ):
   nam = path.realpath(path.join(path.dirname( __file__ ), '..', nam))
   o = outfit(nam, is_multi= core, w= False)
   o.stack(o if doubled else None)
   out = o.to_dict()
   out['size'] = o.size_name()
   return out

# Based on some XML templates and the specs below, we're going to generate families of outfit XML files.
# First things first: what are we supposed to be doing?

class Build:
   ''' By default, build everything in-tree. '''

   def read_template( self, name ):
      return open(f'templates/{name}').read()

   def file_names( self, names ):
      return [f"{n.lower().replace(' ','_')}.xml" for n in names]

   def open_output( self, names ):
      return (open(fn, 'w') for fn in self.file_names(names))

class MesonBuild(Build):
   ''' If Meson invokes generate.py <template path> -o <output paths>, build only the outfit family that matches the command. '''

   def __init__( self, template_path, _, *output_paths ):
      self.template_path = template_path
      self.output_paths = output_paths
      self.template_name = path.basename(template_path)
      self.output_names = [path.basename(p) for p in output_paths]

   def read_template( self, name ):
      if name == self.template_name:
         return open(self.template_path).read()

   def open_output( self, names ):
      if self.file_names(names) == self.output_names:
         for fn in self.output_paths:
            yield open(fn, 'w')

build = MesonBuild(*sys.argv[1:]) if sys.argv[2:3] == ['-o'] else Build()

intify = lambda x: int(x) if round(x) == x else x
# Hopefully we can refactor this once the behaviour is finalized and we're unafraid of merge conflicts. Anyway, back to the show.

def lerpt( t ):
   return lambda x: t[int(math.floor(x*(len(t)*0.99999)))]

def lerp( a, b, ta = 0.0, tb = 1.0 ):
   return lambda x: intify(a + (x-ta) * (b-a) / (tb-ta))

def lerpr( a, b, ta = 0.0, tb = 1.0 ):
   def f( x ):
      res = int(round(lerp(a, b, ta, tb)(x)))
      return 0 if res < 0 else res;
   return f

from math import log, exp

# Does geometric interpolation instead of arithmetic interpolation.
def eerp( a, b, ta = 0.0, tb = 1.0 ):
   return lambda x: intify(round(exp(lerp(log(a), log(b), ta, tb)(x)), 2))

def eerpr( a, b, ta = 0.0, tb = 1.0 ):
   return lambda x: int(round(eerp(a, b, ta, tb)(x)))

class BioOutfit:
   def __init__( self, template, params ):
      self.template = template
      self.params = params
      self.txt = build.read_template(template)

   def run( self, names ):
      files = build.open_output(names)
      for k, (n, f) in enumerate(zip(names, files)):
         x = 1 if len(names) == 1 else k/(len(names)-1)
         p = {k: v(x) if callable(v) else v for k, v in self.params.items()}
         p['name']  = n
         with f:
            txt = self.txt.format(**p)
            txt = txt.replace( '<general>', '<!-- THIS OUTFIT IS AUTOGENERATED. DO NOT EDIT DIRECTLY! -->\n <general>' )
            f.write( txt )

desc = {}
desc['brain'] = N_("The brain is a Soromid bioship's equivalent to core systems in synthetic ships. Possibly the most important organ, the brain provides processing power and allocates energy to the rest of the organism. All brains start off undeveloped, but over time, just like the ships themselves, they grow and improve.")
desc['engine'] = N_("The gene drive is a Soromid bioship's equivalent to engines in synthetic ships. It is charged with moving the organism through space and is even capable of hyperspace travel. All gene drives start off undeveloped, but over time, just like the ships themselves, they grow and improve.")
desc['hull'] = N_("The shell is a Soromid bioship's natural protection, equivalent to hulls of synthetic ships. The shell is responsible both for protecting the organism's internal organs and for creating camouflage to reduce the risk of being detected by hostiles. All shells start off undeveloped, but over time, just like the ships themselves, they grow and improve.")

typename = {}
typename['brain'] = N_("Bioship Brain")
typename['engine'] = N_("Bioship Gene Drive")
typename['hull'] = N_("Bioship Shell")


## BioOutfit generation rules.
## See comments in this directory's meson.build file: these rules must match the build rules.


## Gene Drive recipe:
## Corresponds (when maxed) to some existing engine <ref> mentionned in the comment.
##    "price":        lerpr( <ref>/2.0, <ref> ),
##    "accel":        lerp(  <ref>-15%, <ref> ),
##    "turn":         lerp(  <ref>-15%, <ref> ),
##    "speed":        lerp(  <ref>-15%, <ref> ),

## Cortex recipe:
##    "price":        lerpr(     <S&K>/2, <S&K>),
##    "absorb":       lerpr( <Unicorp>-3, <S&K>-3 ),
##    "armour":       lerp(    <Unicorp>, <S&K> )
##    "cargo":        lerpr(       <S&K>, (<S&K>+<Unicorp>)/2 ),
##    "mass":         <S&K>,
##    "armour":       lerp(    <Unicorp>, <S&K> )

## Cerebrum recipe part 1 (small):
##    "price":        lerpr(  <orion>, <orion>*1.5 ),
##    "mass":         lerpr(  <orion>, <orion>*1.25 ),
##    "shield" :      lerp(   <orion>, <orion>*1.25 ),
##    "shield_regen": lerp(   <orion>, <orion>*1.25 ),

## Cerebrum recipe part 1 (medium/large):
##    "price":        lerpr(  <orion>, <orion>+1/2*(<shield_booster>+<shield capacitor>)),
##    "mass":         lerpr(  <orion>, <orion>+1/2*(<shield_booster>+<shield capacitor>)),
##    "shield":       lerp(   <orion>, <orion>+1/2*<shield capacitor> ),
##    "shield_regen": lerp(   <orion>, <orion>+1/2*<shield_booster> ),

## Cerebrum recipe part 2:
##    "energy" :      lerp(   <orion>, <orion>*1.25 ),
##    "energy_regen": lerp(   <orion>, <orion>*1.25 ),
##    "cpu":          handmade ! (because builtin weapons have no CPU requirements)

# Perlevis Gene Drive   =>  Tricon Zephyr
# Laeviter Gene Drive   =>  Tricon Zephyr x2
# Laevis Gene Drive     =>  Melendez Ox x2
# Mediocris Gene Drive  =>  Tricon Cyclone
# Largus Gene Drive     =>  Tricon Cyclone x2
# Ponderosus Gene Drive =>  Tricon Typhoon
# Grandis Gene Drive    =>  Melendez Mammoth
# Magnus Gene Drive     =>  Tricon Typhoon x2
# Immanis Gene Drive    =>  Nexus Bolt 3000 x2
for siz, nam, db, temp, gfx, output_pref, outputs in [
   ('small', 'tricon_zephyr_engine.xml',     False,   'gene_drive_tricon',    'fast_s',   'Perlevis',    ['I', 'II']          ),
   ('small', 'tricon_zephyr_engine.xml',     True,    'gene_drive_tricon',    'fast_s',   'Laeviter',    ['I', 'II']          ),
   ('small', 'melendez_ox_engine.xml',       True,    'gene_drive_melendez',  'strong_s', 'Laevis',      ['I', 'II']          ),
   ('medium', 'tricon_cyclone_engine.xml',   False,   'gene_drive_tricon',    'fast_m',   'Mediocris',   ['I', 'II', 'III']   ),
   ('medium', 'tricon_cyclone_engine.xml',   True,    'gene_drive_tricon',    'strong_m', 'Largus',      ['I', 'II']          ),
   ('large', 'tricon_typhoon_engine.xml',    False,   'gene_drive_tricon',    'fast_l',   'Ponderosus',  ['I', 'II', 'III']   ),
   ('large', 'melendez_mammoth_engine.xml',  False,   'gene_drive_melendez',  'strong_l', 'Grandis',     ['I', 'II', 'III']   ),
   ('large', 'tricon_typhoon_engine.xml',    True,    'gene_drive_tricon',    'strong_l', 'Magnus',      ['I', 'II', 'III']   ),
   ('large', 'nexus_bolt_3000_engine.xml',   True,    'gene_drive',           'strong_l', 'Immanis',     ['I', 'II', 'III']   ),
]:
   ref = get_outfit_dict(path.join('core_engine', siz, nam), True, db)
   BioOutfit( temp+'.xml.template', {
      'typename':     typename['engine'],
      'size':         ref['size'],
      'price':        lerpr(  ref['price']/2, ref['price'] ),
      'mass':         ref['mass'],
      'desc':         desc['engine'],
      'gfx_store':    lerpt(('organic_engine_'+gfx+'1', 'organic_engine_'+gfx+'2')),
      'accel':        lerpr(  0.85*ref['accel'], ref['accel'] ),
      'speed':        lerpr(  0.85*ref['speed'], ref['speed'] ),
      'turn':         lerpr(  0.85*ref['turn'], ref['turn'] ),
      'fuel':         ref['fuel'],
      'energy_malus': ref['energy_regen_malus'],
      'engine_limit': ref['engine_limit']
   } ).run( [ N_(output_pref+' Gene Drive '+s) for s in outputs ] )


##
# Cortex
#

# Perlevis Cortex    =>  (1) Unicorp_d2      (2) S&K Skirmish Plating
# Laevis Cortex      =>  (1) Unicorp_d2 x2   (2) S&K Skirmish Plating x2
# Mediocris Cortex   =>  (1) Unicorp_d23     (2) S&K Battle Plating
# Largus Cortex      =>  (1) Unicorp_d23 x2  (2) S&K Battle Plating x2
# Ponderosus Cortex  =>  (1) Unicorp_d58     (2) S&K War Plating
# Immanis Cortex     =>  (1) Unicorp_d58 x2  (2) S&K War Plating x2
for pref, nam1, nam2, dbl, gfx, output_pref, outputs in [
   ('small',  'unicorp_d2_light_plating.xml',   'sk_skirmish_plating.xml', False, 't', 'Perlevis',   ['I', 'II']              ),
   ('small',  'unicorp_d2_light_plating.xml',   'sk_skirmish_plating.xml', True,  's', 'Laevis',     ['I', 'II']              ),
   ('medium', 'unicorp_d23_medium_plating.xml', 'sk_battle_plating.xml',   False, 'm', 'Mediocris',  ['I', 'II']              ),
   ('medium', 'unicorp_d23_medium_plating.xml', 'sk_battle_plating.xml',   True,  'l', 'Largus',     ['I', 'II', 'III']       ),
   ('large',  'unicorp_d58_heavy_plating.xml',  'sk_war_plating.xml',      False, 'h', 'Ponderosus', ['I', 'II', 'III', 'IV'] ),
   ('large',  'unicorp_d58_heavy_plating.xml',  'sk_war_plating.xml',      True,  'x', 'Immanis',    ['I', 'II', 'III']       ),
]:
   ref1 = get_outfit_dict(path.join('core_hull', pref, nam1), True, dbl)
   ref2 = get_outfit_dict(path.join('core_hull', pref, nam2), True, dbl)
   BioOutfit( 'cortex.xml.template', {
      'typename':     typename['hull'],
      'size':         pref,
      'price':        lerpr( ref2['price']/2,  ref2['price'] ),
      'mass':         int(ref2['mass']),
      'desc':         desc['hull'],
      'gfx_store':    'organic_hull_' + gfx,
      'cargo':        lerpr( ref2['cargo'],    round((ref1['cargo']+ref2['cargo'])/2.0) ),
      'absorb':       lerpr( ref1['absorb']-3, ref2['absorb']-3 ),
      'armour':       lerpr( ref1['armour'],   ref2['armour'] )
   } ).run( [ N_(output_pref + " Cortex " + s) for s in outputs ] )


##
# Cerebrum
#

# Perleve Cerebrum    =>  Orion_2301
# Laevum Cerebrum     =>  Orion_2301 x2
# Mediocre Cerebrum   =>  Orion_4801
# Largum Cerebrum     =>  Orion_4801 x2
# Ponderosum Cerebrum =>  Orion_8601
# Immane Cerebrum     =>  Orion_8601 x2
for siz, nam, dbl, gfx, output_pref, outputs, cpu in [
   ('small','milspec_orion_2301_core_system.xml',  False, 's1', 'Perleve',    ['I', 'II'],         (     5,     6 )),
   ('small','milspec_orion_2301_core_system.xml',   True, 's2', 'Laevum',     ['I', 'II'],         (    24,    32 )),
   ('medium','milspec_orion_4801_core_system.xml', False, 'm1', 'Mediocre',   ['I', 'II'],         (    48,   100 )),
   ('medium','milspec_orion_4801_core_system.xml',  True, 'm2', 'Largum',     ['I', 'II'],         (   100,   200 )),
   ('large','milspec_orion_8601_core_system.xml',  False, 'l1', 'Ponderosum', ['I', 'II', 'III'],  (   100,   200 )),
   ('large','milspec_orion_8601_core_system.xml',   True, 'l2', 'Immane',     ['I', 'II', 'III'],  ( 370*3, 600*3 ))
]:
   ref = get_outfit_dict(path.join('core_system', siz, nam), True, dbl)

   if gfx[0] == 's':
      price = 0.5*ref['price']
      mass = 0.25*ref['mass']
      shield = 0.25*ref['shield']
      shield_regen = 0.25*ref['shield_regen']
   else:
      shield_capacitor = get_outfit_dict(path.join('structure', 'shield_capacitor_ii'+('i' if gfx[0] ==' l' else '')+'.xml'))
      shield_booster = get_outfit_dict(path.join('structure', ref['size']+'_shield_booster.xml'))

      price = 0.5*(shield_booster['price']+shield_capacitor['price'])
      mass = 0.5*(shield_booster['mass']+shield_capacitor['mass'])
      shield = 0.5*shield_capacitor['shield']
      shield_regen = 0.5*shield_booster['shield_regen']

   BioOutfit( 'cerebrum.xml.template', {
      'price':          lerpr(ref['price'],        ref['price']+price),
      'mass':           lerpr(ref['mass'],         ref['mass']+mass),
      'shield':         lerpr(ref['shield'],       ref['shield']+shield),
      'shield_regen':   lerpr(ref['shield_regen'], ref['shield_regen']+shield_regen),
      'typename':       typename['brain'],
      'desc':           desc['brain'],
      'size':           ref['size'],

      'energy' :        lerp(ref['energy'],        ref['energy']*1.25 ),
      'energy_regen' :  lerp(ref['energy_regen'],  ref['energy_regen']*1.25 ),
      'gfx_store':      'organic_core_'+gfx,
      'cpu':            lerpr(*cpu),
   } ).run( [ N_(output_pref+' Cerebrum '+s) for s in outputs ] )


##
# Weapons
#

# Stinger  =>  Plasma Blaster MK1  &  MK2
extrapol = (0.25, 0.75)
follow = (0.5, 1.0)
BioOutfit( 'weapon.xml.template', {
   'typename':      N_("Bioship Weapon Organ"),
   'size':          'small',
   'mass':          eerpr(    3,    6, *follow),
   'price' :        eerpr( 19e3, 45e3, *extrapol),
   'desc':          N_("The Stinger Organ is able to convert energy into hot plasma that is able to eat easily through shield and armour of opposing ships over time. While not an especially powerful offensive organ, it is prized for its reliability."),
   'gfx_store':     lerpt(('organic_plasma_s1', 'organic_plasma_s2', 'organic_plasma_s3')),
   'specific':      'bolt',
   'gfx':           'plasma',
   'gfx_end':       'plasma2-end',
   'sound':         'bioplasma',
   'spfx_shield':   'Shi-S',
   'spfx_armour':   'Pla-S',
   'lua':           'bioplasma.lua',
   'delay':         eerp(   1.2,  1.4, *follow),
   'speed':         700,
   'range':         eerpr(  700,  800, *extrapol),
   'falloff':       eerpr(  600,  650, *extrapol),
   'energy':        eerpr(    6, 16.5, *follow),
   'trackmin':      0,
   'trackmax':      eerpr( 2000, 3000, *follow),
   'penetrate':     lerpr(    0,   10, *follow),
   'damage':        eerp(    20,   27, *extrapol),
   'extra':         '<swivel>22</swivel>',
} ).run( [
   N_("Stinger Organ I"),
   N_("Stinger Organ II"),
   N_("Stinger Organ III"),
] )

# Talon  =>  Plasma Blaster mk2  &  Plasma Cluster
# We want Talon(0) = Plasma Blaster mk2  and  Talon(3) = Plasma Cluster
# This way, if we admit PB mk2 is size 2 (fighter size) and PCluster is size 3.5 (medium),
# then Talon II is size 3, TIII size 3.5 et TIV size 4.
# A few tweaks for keeping Talon almost the same.
extrapol = (-1.0/3, 2.0/3)

BioOutfit( 'weapon.xml.template', {
   'typename':      N_("Bioship Weapon Organ"),
   'size':          'medium',
   'mass' :         eerpr(    6,    16, -1.0/3, 1.0), # have Talon(4) weight as much as a standard med weapon.
   'price' :        eerpr( 45e3,  95e3, *extrapol),
   'desc':          N_("The Talon Organ is an enlarged and more powerful version of the Stinger Organ. Like its smaller counterpart, is able to convert energy into hot plasma that is able to eat easily through shield and armour of opposing ships. The hot plasma is able to cling to ship's shields and hulls dealing continuous damage after impact."),
   'gfx_store':     'organic_plasma_l',
   'specific':      'bolt',
   'gfx':           'plasma2',
   'gfx_end':       'plasma2-end',
   'sound':         'bioplasma',
   'spfx_shield':   'Shi-S',
   'spfx_armour':   'Pla-S2',
   'lua':           'bioplasma2.lua',
   'delay':         eerp(   1.4,  2.2, *extrapol),
   'speed' :        700,
   #
   'range' :        eerpr(  800, 1200, *extrapol),
   'falloff':       eerpr(  650, 1000, *extrapol),
   'energy':        eerpr( 16.5,   66, *extrapol),
   'trackmin':      lerpr(    0, 1000, *extrapol),
   'trackmax':      eerpr( 3000, 5000, *extrapol), # 5000 was 6000
   'penetrate':     lerpr(   10,   42, *extrapol),
   'damage':        eerp(    27,   68, *extrapol),
   #'range' :       lerp(  1000, 1400 ),
   #'falloff':      lerp(   900, 1200 ),
   #'energy':       lerp(    46,   58 ),
   #'trackmin':     lerp(  1500, 2000 ),
   #'trackmax':     lerp(  4500, 6000 ),
   #'penetrate':    lerpr(   35,   50 ),
   #'damage':       lerp(    35,   47 ),
   'extra':         '<swivel>22</swivel>',
} ).run( [
   N_("Talon Organ I"),
   N_("Talon Organ II"),
   N_("Talon Organ III"),
   N_("Talon Organ IV"),
] )

# Tentacle Organ  =>  .. complicated ..
#
# By examining plasma_turret mk2 and laser_turret mk2, we could guess the following:
#  - laser has +20% range +20% falloff +20% tackmin/max +12.5% penetrate over plasma +1/16 more costly
#  - plasma has +40% delay +40% energy +10% damage over laser
#  - they have same mass and CPU usage
#  - speed is constant, 800 in the case of plasma.
#
# Now we can guess what a heavy plasma turret and turboplasma turret would look like.
# There only remains to interpolate/extrapolate as we did for stinger.

# heavy plasma = tentacle 2  turboplasma = tentacle 4
extrapol = (1.0/3.0, 1.0)
extrapol_bias = (1.0/3, 1.0)#+1.0/6)

BioOutfit( 'weapon.xml.template', {
   'typename':      N_("Bioship Weapon Organ"),
   'size':          'large',
   'mass':          eerpr(         36,          75, *extrapol),
   'price':         eerpr(320e3*16/17, 700e3*16/17, *extrapol),
   'desc':          N_("The Tentacle Organ has the distinction of being the only fully rotating organic weapon while boasting a fully developed power output that is hard to beat with conventional weaponry found throughout the galaxy. The large globs of hot plasma it launches can corrode through seemingly impregnable armours, seeping into and melting ships from the inside upon contact."),
   'gfx_store':     'organic_plasma_t',
   'specific':      'turret bolt',
   'gfx':           'plasma',
   'gfx_end':       'plasma2-end',
   'sound':         'bioplasma_large',
   'spfx_shield':   'Shi-M',
   'spfx_armour':   'Pla-M',
   'lua':           'bioplasma.lua',
   'delay':         eerp(    1.4*1.4,  1.4*1.45, *extrapol),
   'speed' :        700,
   #
   'range' :        eerpr(  1750/1.2,  2400/1.2, *extrapol),
   'falloff':       eerpr(  1400/1.2,  1800/1.2, *extrapol),
   'energy':        eerpr(    49*1.4,   174*1.4, *extrapol_bias),
   'trackmin':      eerpr(  2400/1.2,  6000/1.2, *extrapol),
   'trackmax':      eerpr(  9600/1.2, 24000/1.2, *extrapol),
   'penetrate':     lerpr(  62/1.125,  80/1.125, *extrapol),
   'damage':        eerp(     48*1.1,    70*1.1, *extrapol),
   #'range' :       lerp(       1800,  2000 ),
   #'falloff':      lerp(       1500,  1800 ),
   #'energy':       lerp(        155,   190 ),
   #'trackmin':     lerp(       3000,  4000 ),
   #'trackmax':     lerp(      16000, 20000 ),
   #'penetrate':    lerpr(        70,   100 ),
   #'damage':       lerp(         51,    65 ),
   'extra':         '<swivel>22</swivel>',
} ).run( [
   N_("Tentacle Organ I"),
   N_("Tentacle Organ II"),
   N_("Tentacle Organ III"),
   N_("Tentacle Organ IV")
] )



#!/usr/bin/env python3

from sys import stderr, stdin
from outfit import LOWER_BETTER
from arg_to_obj import arg_to_naev_obj

def transpose( M ):
   N = max(map(len, M))
   M = [t + ['']*(N-len(t)) for t in M]
   return list(zip(*(tuple(M))))

getfloat = lambda s: float(s.split('/')[0])

def keyfunc( s ):
   def key( o ):
      try:
         return getfloat(o[s])
      except:
         return None
   return key

def nfmt( n ):
   f = float(n)
   if f == round(f):
      f = int(f)
   return str(f)

def main( args, gith = False, colour = False, term = False, noext = False,
      sortit = False, autostack = False, combine = False, good = False ):
   if not args:
      return 0

   rang = dict()
   L = []
   names = []
   acc = []

   for o in arg_to_naev_obj(args, combine, autostack, good):
      d = o.to_dict()
      for k, v in d.items():
         if type(v) == type(()):
            (val, _) = v
            s = nfmt(v[0]) + '/' + nfmt(v[1])
         elif type(v) in {type(1), type(1.0)}:
            val = v
            s = nfmt(v)
         else:
            continue

         d[k] = s
         if k not in rang:
            rang[k] = (val, val)
            acc.append(k)
         else:
            (mi, ma) = rang[k]
            rang[k] = (min(mi, val), max(ma, val))

      names.append(o.shortname())
      L.append(d)
   stderr.write('\n')

   for i, t in rang.items():
      (m, M) = t
      if i in LOWER_BETTER:
         rang[i] = (M, m)
      if noext:
         rang[i] = (None, None)

   if sortit:
      for i, o in enumerate(L):
         o['name'] = names[i]
      L.sort(key = keyfunc(sortit))
      for i, o in enumerate(L):
         names[i] = o['name']

   Res = [[k]+[l[k] if k in l else '' for l in L] for k in acc]
   names = [''] + names
   length = [4] * len(names) # :--- at least 3 - required

   termit_rule = lambda s: s
   if colour:
      Sep, SepAlt, Lm, Rm, LM, RM = '\033[30;1m|\033[0m', '\033[34m|\033[0m', '\033[31m', '\033[37m', '\033[32m', '\033[37m'
      mk_pad = lambda i, n: '\033[34m' + n*'-' + '\033[0m'
      leng = lambda x: len(x) - (10 if x != '' and x[0]=='\033' else 0)
      if term:
         # Could also use it to manage gith and colour.
         from slst import Slst
         termit = Slst([('|',"\N{BOX DRAWINGS LIGHT VERTICAL}")])
         (Sep, SepAlt, Lm, Rm, LM, RM) = map(termit, (Sep, SepAlt, Lm, Rm, LM, RM))
         termit_rule = -termit + [
            ('\033[34m', ), ('\033[0m', ),            # uncolour
            ('| -', '|--'), ('- |',  '--|'),
            ('-|-',  "-\N{BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL}-" ),
            ('|-',   "\N{BOX DRAWINGS LIGHT DOWN AND RIGHT}-"           ),
            ('-|',   "-\N{BOX DRAWINGS LIGHT VERTICAL AND LEFT}"        ),
            ('^|',   "\N{BOX DRAWINGS LIGHT VERTICAL AND RIGHT}"        ),
            ('-',    "\N{BOX DRAWINGS LIGHT HORIZONTAL}"                ),
            ('', '\033[34m', 1), ('', '\033[0m', -1)  # recolour whole line
         ]
   else:
      Sep, SepAlt, Lm, Rm, LM, RM = '|', '|', '_', '_', '**', '**'
      mk_pad = lambda i, n: '-'*(n-1)+('-' if i == 0 else ':')
      leng = len

   if colour:
      head = transpose([s.split(' ') for s in names])
   else:
      head = [names]

   if not gith:
      for r in Res + head:
         length = [max(n, len(s)) for (n, s) in zip(length, r)]

   mklinsep = lambda sep: lambda L:sep+' '+(' '+sep+' ').join(L)+' '+sep
   mklinalt = lambda alt: mklinsep(SepAlt) if alt else mklinsep(Sep)
   mklin = mklinalt(False)
   fmt = lambda t: (t[1]-leng(t[0]))*' '+t[0]
   lfmt = lambda t: t[0]+(t[1]-leng(t[0]))*' '

   def emph( s, m ):
      (mi, ma) = m
      if s == '':
         return '_'
      elif mi != ma:
         if getfloat(s) == mi:
            return Lm + s + Rm
         elif getfloat(s) == ma:
            return LM + s + RM
      return s

   print()
   if term:
      toprule = (length[0]+3)*' ' + mklinalt(True)([mk_pad(i, n) for i, n in enumerate(length[1:])])
      print((termit_rule + [ ('\033[34m', '\033[30;1m'),
         ("\N{BOX DRAWINGS LIGHT VERTICAL AND LEFT}","\N{BOX DRAWINGS LIGHT DOWN AND LEFT}"),
         ("\N{BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL}","\N{BOX DRAWINGS LIGHT DOWN AND HORIZONTAL}")
      ])(toprule))

   for t in head:
      acc = mklin(map(fmt, zip(t, length)))
      if colour:
         acc = ' '+acc[len(Sep):]
      print(acc)
   print(termit_rule(mklinalt(True)([mk_pad(i, n) for i, n in enumerate(length)])))

   count = 0
   for r in Res:
      r = [r[0].replace('_', ' ')] + [emph(k, rang[r[0]]) for k in r[1:]]
      print(mklinalt(count%3 == 2)([fmt(x) if i>0 else lfmt(x) for i, x in enumerate(zip(r, length))]))
      count += 1
   return 0

if __name__ == '__main__':
   import argparse

   parser = argparse.ArgumentParser(
      usage = ' %(prog)s  [-g|-c] [-n] [(-s SORT) | -S]  [filename ...]',
      formatter_class = argparse.RawTextHelpFormatter,
      description = 'By default, outputs text aligned markdown table comparing the outfits respective values.',
      epilog = '\nTypical usage (from naev root dir) :\n > ./utils/outfits/outfits2md.py dat/outfits/core_system/small/*.xml -C -t | less -RS'
   )
   parser.add_argument('-g', '--github', action = 'store_true', help = 'unaligned (therefore smaller) valid github md, for use in posts.')
   parser.add_argument('-c', '--colour', action = 'store_true', help = 'coloured terminal output. You can pipe to "less -RS" if it is too wide.')
   parser.add_argument('-t', '--term', action = 'store_true', help = 'coloured terminal output (TODO: with extended table characters).')
   parser.add_argument('-n', '--nomax', action = 'store_true', help = 'Do not emphasize min/max values.' )
   parser.add_argument('-s', '--sort', help = 'inputs are sorted by their SORT key.')
   parser.add_argument('-S', '--sortbymass', action = 'store_true', help = 'Like -s mass.' )
   parser.add_argument('-A', '--autostack', action = 'store_true', help = 'Outfits are presented both alone and auto-stacked.')
   parser.add_argument('-C', '--combinations', action = 'store_true', help = 'Does all the combinations.' )
   parser.add_argument('-G', '--good', action = 'store_true', help = "Like -C, but good comb. only: don't comb when pri smaller." )
   parser.add_argument('-f', '--files', action = 'store_true', help = 'read file list on stdin. Applies when no args.')
   parser.add_argument('filename', nargs = '*', help = """An outfit with ".xml" or ".mvx" extension, else will be ignored.
Can also be two outfits separated by \'+\', or an outfit prefixed
with \'2x\' (or: \'1x\') or suffixed with \'x2\' (or: \'x1\').""")

   args = parser.parse_args()
   if args.term and args.colour:
      print('Warning: -c subsumed by -t', file = stderr, flush = True)

   if args.github and (args.colour or args.term):
      print('Warning: Ignored conflicting: -g and -' +
         ('t' if args.term else 'c'), file = stderr, flush = True)
      args.github = args.colour = args.term = False

   if args.autostack and args.good:
      print('Warning: -A is subsumed by -G', file = stderr, flush = True)
      args.autostack = False
   elif args.autostack and args.combinations:
      print('Warning: -A is subsumed by -C', file = stderr, flush = True)
      args.autostack = False

   if args.good and args.combinations:
      print('Warning: -C is subsumed by -G', file = stderr, flush = True)

   ign = [f for f in args.filename if not f.endswith('.xml') and not f.endswith('.mvx')]
   if ign:
      print('Ignored: "'+'", "'.join(ign)+'"', file = stderr, flush = True)

   sortby = args.sort or (args.sortbymass and 'mass') or False

   if sortby:
      print('sorted by "'+str(sortby)+'"', file = stderr, flush = True)

   if args.files or not args.filename:
      args.filename += [l.strip() for l in stdin.readlines()]

   main([f for f in args.filename if f not in ign],
      args.github, args.colour or args.term, args.term, args.nomax, sortby,
      args.autostack, args.combinations or args.good, args.good)
else:
   raise Exception('This module is only intended to be used as main.')



#!/usr/bin/env python3
# Standalone Valgrind server for Naev.
#
# Usage:
#   1. Run this script in Terminal A:         ./naev_valgrind.py
#   2. Attach the debugger in Terminal B:     WITHVALGRIND=true ./naev.py
#      (Alternatively, use 'vgdb' or 'gdb' directly as instructed by this script).
#
#   Pro-tip: To see game stdout in the debugger terminal, run `tty` in that
#            terminal to get its device path (e.g. /dev/pts/1), then start
#            this server with: ./naev_valgrind.py > /dev/pts/1 2>&1
#
# Environment Variables:
#   VG_TRACE_CHILDREN=true  Enable --trace-children=yes in Valgrind.
#   VG_LOGFILE=filename     Specify a log file for Valgrind output (default: build_root/naev_valgrind.log).
#   WITHDEBUGGER=true       Enable Valgrind server mode (vgdb) for GDB attachment.
#   VG_SUPPRESSIONS=file    Path to a Valgrind suppression (.supp) file (default: @naev_supp@).
#   VG_ARGS="--flags"       Additional raw arguments to pass to Valgrind.
#
# Automated Settings (Applied by script):
#   ALSOFT_LOGLEVEL      Set to 2 (debug) or 3 (paranoid) to help with OpenAL debugging.
#   ALSOFT_TRAP_AL_ERROR Set to 1 (paranoid only) to catch OpenAL errors immediately.
#   RUST_BACKTRACE       Always set to 1 to provide detailed Rust error reports.
#   ASAN_OPTIONS         Always set to halt_on_error=1 for precise memory checking.

import os
import subprocess
import shutil
import sys
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def env_bool(name: str, default: bool) -> bool:
   v = os.getenv(name)
   if v is None:
      return default
   v = v.strip().lower()
   return v in ("1", "true", "yes", "y", "on")

source_root = "@source_root@"
build_root = "@build_root@"
debug_paranoid = "@debug_paranoid@"
debug = "@debug@"
naev_bin = "@naev_bin@"
zip_overlay = "@zip_overlay@"

# Optional VG knobs
VG_TRACE_CHILDREN = env_bool("VG_TRACE_CHILDREN", False)
VG_LOGFILE = os.getenv("VG_LOGFILE", os.path.join(build_root, "naev_valgrind.log"))
WITHDEBUGGER = env_bool("WITHDEBUGGER", False)

# Default suppressions to naev.supp in the utils/build directory
naev_supp = os.path.join(source_root, "utils", "build", "naev.supp")
VG_SUPPRESSIONS = [naev_supp]
if os.getenv("VG_SUPPRESSIONS"):
   VG_SUPPRESSIONS.extend(os.getenv("VG_SUPPRESSIONS").split(','))

VG_ARGS = os.getenv("VG_ARGS", "")

MESON = [sys.executable, os.path.join(source_root, "meson.py")]

def wrapper(*args):
   if not os.path.exists(os.path.join(source_root, "meson.py")):
      logger.error(f"Error: meson.py not found at {source_root}. Check your source tree.")
      return

   if debug_paranoid == "True":
      os.environ["ALSOFT_LOGLEVEL"] = "3"
      os.environ["ALSOFT_TRAP_AL_ERROR"] = "1"
   elif debug == "True":
      os.environ["ALSOFT_LOGLEVEL"] = "2"
   os.environ["RUST_BACKTRACE"] = "1"
   os.environ["ASAN_OPTIONS"] = "halt_on_error=1"

   if not shutil.which("valgrind"):
      logger.error("Error: valgrind is not installed or not in PATH.")
      return

   if not WITHDEBUGGER:
      logger.info("Valgrind Standalone Mode: Running game directly under Valgrind...")
   else:
      logger.info("Valgrind Server Mode: Waiting for GDB connection via vgdb...")

   # Define an isolated pipe prefix in the build directory to ensure client/server find each other.
   vgdb_prefix = os.path.join(build_root, ".vgdb-pipe")

   valgrind_command = [
      "valgrind",
      "--leak-check=full",
      "--show-leak-kinds=all",
      "--track-origins=yes",
      "--num-callers=100",
      "--error-limit=no",
      "--vex-guest-max-insns=25", # Workaround for temporary storage exhaustion
   ]

   if WITHDEBUGGER:
      valgrind_command += [
         "--vgdb=yes",
         "--vgdb-error=0",
         f"--vgdb-prefix={vgdb_prefix}",
      ]

   for suppression in VG_SUPPRESSIONS:
      if os.path.exists(suppression):
         valgrind_command += [f"--suppressions={suppression}"]
      else:
         logger.warning(f"Suppression file not found: {suppression}")

   if VG_ARGS:
      import shlex
      valgrind_command += shlex.split(VG_ARGS)

   if VG_TRACE_CHILDREN:
      valgrind_command += ["--trace-children=yes"]
   if VG_LOGFILE:
      valgrind_command += [f"--log-file={VG_LOGFILE}"]

   full_command = MESON + ["devenv", "-C", build_root] + valgrind_command + list(args)

   try:
      p = subprocess.Popen(full_command)
      p.wait()
   except KeyboardInterrupt:
      logger.info("\nCtrl+C detected, shutting down Valgrind...")
      p.terminate()
      try:
         p.wait(timeout=5)
      except subprocess.TimeoutExpired:
         p.kill()
   except Exception as e:
      logger.error(f"Failed to execute Valgrind: {e}")
   finally:
      if sys.platform != "win32":
         os.system("stty sane 2>/dev/null")

# Build steps
subprocess.run(MESON + ["compile", "-C", build_root, "naev-gmo"], check=False)
os.makedirs(os.path.join(build_root, "dat/gettext"), exist_ok=True)
po_root = os.path.join(build_root, "po")
if os.path.isdir(po_root):
   for mo_name in os.listdir(po_root):
      mo_path = os.path.join(po_root, mo_name)
      if os.path.isdir(mo_path):
         dest_dir = os.path.join(build_root, "dat/gettext", mo_name)
         shutil.copytree(mo_path, dest_dir, dirs_exist_ok=True)

wrapper(
   naev_bin,
   "-d", zip_overlay,
   "-d", os.path.join(source_root, "dat"),
   "-d", os.path.join(source_root, "assets"),
   "-d", os.path.join(build_root, "dat"),
   "-d", source_root,
   *sys.argv[1:],
)



#!/usr/bin/env python3

from sys import argv, stderr, stdout, exit
from os import path
from pathlib import Path

if __name__ != '__main__':
   raise Exception('I am a teapot')

my_name, *args = argv

if '-h' in args or len(args) < 2:
   stderr.write(
      'usage: ' + my_name + '<src_path> <dst_path> <file.xml>...\n'
      'for each <file.xml>, produces an uncommented version that is written to a file\n'
      'that has the same position in <dst_path> that <file.xml> has in <src_path>.\n'
      'The output filename is written to stdout.\n'
      '\n'
      'This script is not intended to be used as is (but can),\n'
      'it is a subroutine of xml_format.\n'
   )
   exit(0)

src, dst, *args = args

for arg in args:
   if (rel := path.relpath(path.realpath(arg), src))[:2] == '..':
      stderr.write('\033[31merror:\033[0m file ' + repr(arg) + ' not in ' + repr(src) + '\n')
      exit(1)

   oarg = path.join(dst, rel)
   Path(path.split(oarg)[0]).mkdir(parents= True, exist_ok= True)
   with open(arg, 'r') as fpr, open(oarg, 'w') as fpw:
      s = fpr.read()
      bef, *aft = s.split('<!--', 1)
      while aft:
         fpw.write(bef.rstrip(' '))
         s, = aft
         bef, *aft = s.split('-->', 1)
         if not aft:
            stderr.write('\033[31merror:\033[0m "<!--" not followed by "-->"')
            exit(-1)
         s, = aft
         s = s.lstrip(' ')
         bef, *aft = s.split('<!--', 1)
      fpw.write(s)
   stdout.write(oarg + '\n')



#!/usr/bin/env python3

import os
import subprocess
import sys
import logging
from typing import Optional

# Configure logging to output to stderr
logging.basicConfig(level=logging.INFO, stream=sys.stderr)


def get_commit_sha(source_root: str) -> Optional[str]:
   """Return the short git commit SHA, or None if unavailable."""
   try:
      proc = subprocess.run(
         ["git", "-C", source_root, "rev-parse", "--short", "HEAD"],
         capture_output=True,
         text=True,
         check=True,
      )
      return proc.stdout.strip()
   except Exception:
      return None


def is_dirty(source_root: str) -> bool:
   """Return True if the git repo has uncommitted changes."""
   try:
      proc = subprocess.run(
         ["git", "-C", source_root, "status", "--porcelain"],
         capture_output=True,
         text=True,
         check=True,
      )
      return bool(proc.stdout.strip())
   except Exception:
      return False


def is_tagged_release(source_root: str) -> bool:
   """Return True if HEAD is exactly at a tag."""
   try:
      subprocess.run(
         ["git", "-C", source_root, "describe", "--exact-match", "--tags", "HEAD"],
         capture_output=True,
         text=True,
         check=True,
      )
      return True
   except subprocess.CalledProcessError:
      return False
   except Exception:
      return False


def get_version(source_root: str, base_version: str) -> str:
   """
   Determine the project version:
      1. If HEAD is exactly at a tag and the repo is clean, use base_version.
         (Assumes base_version has been set to the tag by Meson.)
      2. Otherwise, build a dev version: base_version + git SHA [+dirty] if .git exists.
      3. If no .git, but dat/VERSION exists, read and reuse it.
      4. If nothing else works, fall back to base_version + "+dev".
      The resolved version is always written to dat/VERSION.
   """
   version_file = os.path.join(source_root, "dat", "VERSION")

   version = None

   if os.path.isdir(os.path.join(source_root, ".git")):
      if is_tagged_release(source_root) and not is_dirty(source_root):
         # Exact tag, clean repo
         version = base_version
      else:
         # Dev build: append sha and maybe dirty
         version = base_version
         sha = get_commit_sha(source_root)
         if sha:
            version += f"+g{sha}"
         if is_dirty(source_root):
            version += ".dirty"
   elif os.path.isfile(version_file):
      # No git, but VERSION file exists (tarball or source package)
      try:
         with open(version_file, "r") as f:
            version = f.read().strip()
      except Exception as e:
         logging.warning(f"Could not read existing VERSION file: {e}")

   if not version:
      # Fallback: just use the base project version
      version = f"{base_version}+dev"

   # Write the version to VERSION file
   try:
      os.makedirs(os.path.dirname(version_file), exist_ok=True)
      with open(version_file, "w") as f:
         f.write(version)
   except Exception as e:
      logging.warning(f"Could not write VERSION file: {e}")

   return version


if __name__ == "__main__":
   source_root = os.getenv("MESON_SOURCE_ROOT")
   if not source_root:
      logging.error("This script is intended to be run by Meson.")
      sys.exit(1)
   if len(sys.argv) < 2:
      logging.error("Usage: gen_version.py <project_version>")
      sys.exit(1)

   base_version = sys.argv[1]
   version = get_version(source_root, base_version)
   logging.info(f"Version retrieved: {version}")
   print(version)



#!/usr/bin/env python3
import os
import sys
import shutil
import logging

# Configure logging to stderr and stdout
logging.basicConfig(level=logging.INFO, stream=sys.stderr)

def copy_file(source_file, dest_dir):
   try:
      os.makedirs(dest_dir, exist_ok=True)
      shutil.copy(source_file, os.path.join(dest_dir, os.path.basename(source_file)))
      logging.info(f"File '{source_file}' copied to '{dest_dir}'.")
   except Exception as e:
      logging.error(f"Error copying file: {e}")
      sys.exit(1)

if __name__ == "__main__":
   if len(sys.argv) < 3:
      logging.error("Usage: python3 copydir.py <destination_directory> <file1> <file2> ...")
      sys.exit(1)

   outdir = sys.argv[1]
   for f in sys.argv[2:]:
      copy_file(f, outdir)



#!/usr/bin/env python3
import re
import sys
import shutil
import subprocess
from multiprocessing import Pool

def parse_includes( linesin ):
   lines = linesin.copy()
   i = 0
   while i < len(lines):
      lines[i] = lines[i].replace("%d","0");
      lines[i] = lines[i].replace("%f","0.0");
      lines[i] = lines[i].replace("%s","A");
      m = re.match(r'#include\s+"([^"]+)"', lines[i].strip())
      if m:
         with open("dat/glsl/" + m.group(1)) as f:
            contents = f.read()
         lines = lines[:i] + parse_includes(contents.splitlines()) + lines[i+1:]
      i += 1
   return lines

def preprocess( shader, shadertype ):
   lines = shader.splitlines()

   # Recursively process includes
   lines = parse_includes( lines )

   # Detect if it's a love shader, must be run after #includes are tested
   islove = False
   ispp = False
   for l in lines:
      if not islove and not ispp:
         # Pretty bad detection, but no false positives... for now
         if shadertype=="frag" and re.match(r"vec4 effect\(\s*vec4", l.strip()):
            islove = True
         elif shadertype=="vert" and re.match(r"vec4 position\(\s*mat4", l.strip()):
            islove = True
         elif shadertype=="frag" and re.match(r"vec4 effect\(\s*sampler2D", l.strip()):
            ispp = True

   # Love shaders need some additional voodoo
   prepend = "#version 330 core\n"
   if ispp: # special post-processing shaders
      prepend += """
uniform sampler2D MainTex;
uniform vec4 love_ScreenSize;
in vec4 VaryingTexCoord;
out vec4 colour_out;

vec4 effect( sampler2D tex, vec2 texcoord, vec2 pixcoord );

void main (void)
{
   colour_out = effect( MainTex, VaryingTexCoord.st, vec2(VaryingTexCoord.s,1.0-VaryingTexCoord.t) * love_ScreenSize.xy );
}
"""
   if islove:
      prepend += """
#define _LOVE
// Syntax sugar
#define Image         sampler2D
#define ArrayImage     sampler2DArray
#define VolumeImage    sampler3D
#define Texel         texture
#define love_PixelColor colour_out
#define love_PixelColour colour_out
#define love_Position   gl_Position
#define love_PixelCoord love_getPixelCoord()

// Uniforms shared by pixel and vertex shaders
uniform mat4 ViewSpaceFromLocal;
uniform mat4 ClipSpaceFromView;
uniform mat4 ClipSpaceFromLocal;
uniform mat3 ViewNormalFromLocal;
uniform vec4 love_ScreenSize;
uniform vec4 ConstantColour = vec4(1.0);

// Compatibility
#define TransformMatrix          ViewSpaceFromLocal
#define ProjectionMatrix         ClipSpaceFromView
#define TransformProjectionMatrix   ClipSpaceFromLocal
#define NormalMatrix            ViewNormalFromLocal
"""
      if shadertype=="frag":
         prepend += """
#define PIXEL
uniform sampler2D MainTex;

in vec4 VaryingTexCoord;
in vec4 VaryingColour;
in vec2 VaryingPosition;
out vec4 colour_out;

vec2 love_getPixelCoord() {
   vec2 uv = love_ScreenSize.xy * (0.5*VaryingPosition+0.5);
   uv.y = uv.y * love_ScreenSize.z + love_ScreenSize.w;
   return uv;
}

vec4 effect( vec4 vcolour, Image tex, vec2 texcoord, vec2 pixcoord );

void main(void) {
   love_PixelColour = effect( VaryingColour, MainTex, VaryingTexCoord.st, love_PixelCoord );
}
"""
      else:
         prepend += """
#define VERTEX
in vec4 VertexPosition;
in vec4 VertexTexCoord;
in vec4 VertexColour;

out vec4 VaryingTexCoord;
out vec4 VaryingColour;
out vec2 VaryingPosition;

vec4 position( mat4 clipSpaceFromLocal, vec4 localPosition );

void main(void) {
   VaryingTexCoord  = VertexTexCoord;
   VaryingTexCoord.y= 1.0 - VaryingTexCoord.y;
   VaryingTexCoord  = ViewSpaceFromLocal * VaryingTexCoord;
   VaryingColour    = ConstantColour;
   love_Position   = position( ClipSpaceFromLocal, VertexPosition );
   VaryingPosition  = love_Position.xy;
}
"""
   return prepend+'\n'.join(lines)

def glslvalidate( filename ):
   if filename.endswith(".vert"):
      s = "vert"
   elif filename.endswith(".frag"):
      s = "frag"
   else:
      return

   with open(filename) as f:
      data = preprocess( f.read(), s )

   if "// novalidation" in data:
      return 0, "", ""

   args = ["glslangValidator", "--stdin", "-S", s, "--enhanced-msgs"]
   ret = subprocess.run( args, capture_output=True, input=bytes(data,'utf-8') )
   return ret.returncode, ret.stdout, data

if __name__ == "__main__":
   if shutil.which("glslangValidator") is None:
      sys.exit(0) # Not installed so skip

   filelist = sys.argv[1:]
   filelist.sort()
   with Pool() as pool:
      retlist = pool.map( glslvalidate, filelist )
   err = 0
   for i,r in enumerate(retlist):
      if r[0]!=0:
         err += r[0]
         # only write to stdeout in class of error for less spam
         lines = str(r[1],'utf-8').strip().splitlines()
         sys.stdout.buffer.write(bytes("\n",'utf-8'))
         # TODO something better than inserting all code
         for j,l in enumerate(r[2].splitlines()):
            sys.stdout.buffer.write(bytes(f"{j:04d}:"+l+'\n','utf-8'))
         lines[0] = filelist[i]
         for l in lines:
            sys.stdout.buffer.write( bytes(l+'\n','utf-8') )
   sys.exit( err )



from collections import namedtuple
from dataclasses import dataclass
import glob
import math
import numpy as np
import os
import scipy.sparse as sp
import xml.etree.ElementTree as ET


Asset = namedtuple('Asset', 'x y presences')
Faction = namedtuple('Faction', 'generators invisible lane_base_cost lane_length_per_presence useshiddenjumps')
FactionPresence = namedtuple('FactionPresence', 'facname presence')
Generator = namedtuple('Generator', 'faction weight')

@dataclass
class Presence:
   base: float = 0.
   bonus: float = 0.
   range: int = 0.

   @property
   def value(self):
      return self.base + self.bonus

   def __bool__(self):
      return bool(self.base or self.bonus)

   def __add__(self, rhs):
      return Presence(max(self.base, rhs.base), self.bonus + rhs.bonus)

   def __mul__(self, scalar):
      return Presence(self.base * scalar, self.bonus * scalar)

   def positive_part(self):
      return Presence(max(0., self.base), max(0., self.bonus))


def readFactions(path):
   '''Returns a dictionary of Faction values by name. '''
   full = {}

   for fileName in glob.glob(os.path.join(path, '*.xml')):
      tree = ET.parse(path+fileName)
      root = tree.getroot()
      name = root.attrib['name']

      full[name] = Faction(
         [Generator(gen.text, float(gen.attrib['weight'])) for gen in root.findall('generator')],
         bool(root.find('invisible')),
         float(root.findtext('lane_base_cost', 0)),
         float(root.findtext('lane_length_per_presence', 0)),
         bool(root.find('useshiddenjumps')),
      )
   return {n: f for n, f in full.items() if f.lane_length_per_presence and not f.invisible}

def parse_pos(pos):
   if pos is None:
      return (None, None)
   elif 'x' in pos.attrib:
      return ( float(pos.attrib['x']), float(pos.attrib['y']) )
   else:
      return ( float(pos.findtext('x')), float(pos.findtext('y')) )

def readAssets(path):
   '''Returns a dictionary of Asset values by name. '''
   assets = {}

   for fileName in glob.glob(os.path.join(path, '*.xml')):
      tree = ET.parse((path+fileName))
      root = tree.getroot()
      name = root.attrib['name']
      x, y = parse_pos(root.find('pos'))

      assets[name] = Asset(x, y, presences=[
         FactionPresence(
            presence.findtext('faction'),
            Presence(
               base=float(presence.findtext('base', 0)),
               bonus=float(presence.findtext('bonus', 0)),
               range=int(presence.findtext('range', 0)),
            ),
         )
         for presence in root.findall('presence')
      ])

   return assets


class Systems:
   '''Readable representation of the systems.'''
   def __init__( self ):
      path = '../../dat/ssys/'
      assets  = readAssets( '../../dat/spob/' )
      vassets  = readAssets( '../../dat/spob_virtual/' )
      facinfo = readFactions( '../../dat/factions/' )

      self.facnames = list(facinfo)
      self.lane_base_cost   = {i: facinfo[facname].lane_base_cost         for i, facname in enumerate(self.facnames)}
      self.dist_per_presence = {i: facinfo[facname].lane_length_per_presence for i, facname in enumerate(self.facnames)}
      factions = {name: i for (i, name) in enumerate(self.facnames)}

      self.sysdict = {} # This dico will give index of systems
      self.sysnames = [] # List giving the invert of self.sysdict
      self.nodess = [] # This is a list of nodes in systems

      self.jpnames = [] # List of jump points
      hjnames = [] # List of hidden jump points
      self.jpdicts = [] # List of dict (invert of self.jpnames)

      self.autoposs = []
      self.radius = [] # List of radius of systems
      self.xlist = []
      self.ylist = []
      self.presences = [] # List of presences in systems

      self.loc2globs = [] # Gives the global numerotation from local one for nodes
      self.jp2locs = [] # Gives the local numerotation from jump number
      self.ass2g   = [] # Gives the global numerotation from asset one
      self.g2ass   = [] # Gives asset numerotation from global one
      self.g2sys   = [] # Gives the system from global numerotation

      self.sysass = [] # Assets names per system
      sysvass = [] # Virtual asset names (conferring faction presence or anti-presence)

      self.presass = [] # List of presences in assets
      self.factass = [] # Factions in assets. (I'm sorry for all these asses)

      i = 0 # No of system
      nglob = 0 # Global nb of nodes
      nasg = 0 # Global nb of assest
      for fileName in sorted(os.listdir(path)):
         tree = ET.parse((path+fileName))
         root = tree.getroot()

         name = root.attrib['name']
         self.sysdict[name] = i
         self.sysnames.append(name)

         x, y = parse_pos(root.find('pos'))
         self.xlist.append( x )
         self.ylist.append( y )
         self.radius.append( float(root.findtext('general/radius')) )

         # Store list of nodes (without jumps)
         nodes = []
         loc2glob = []
         sysas = []
         sysvas = []
         nass = 0

         for pnt in root.findall('spobs/spob'):
            asname = pnt.text
            info = assets[asname]
            assert len(info.presences) <= 1, f'Real asset {asname} unexpectedly has multiple presence elements.'
            for facname, facpres in info.presences:
               if not facpres:
                  continue
               sysas.append(asname)
               nodes.append( (info.x, info.y) )
               loc2glob.append(nglob)
               self.ass2g.append(nglob)
               self.g2ass.append(nasg)
               self.g2sys.append(i)
               self.presass.append(facpres)
               self.factass.append(factions.get(facname, -1))
               nglob += 1
               nass += 1
               nasg += 1
         for virt in root.findall('spobs/spob_virtual'):
            sysvas.append(virt.text)

         presence = [Presence() for _ in factions]
         self.presences.append(presence)
         self.sysass.append(sysas)
         sysvass.append(sysvas)


         # Store jump points.
         jpname = []
         hjname = []
         jpdict = {}
         autopos = []
         jp2loc = []
         jumps = root.find('jumps')
         jplist = jumps.findall('jump')
         jjj = 0
         for jpt in jplist:
            if jpt.find('hidden') is not None:
               hjname.append(jpt.attrib['target'])
               continue

            if jpt.find('exitonly') is not None:
               continue

            nodes.append( parse_pos( jpt.find('pos') ) )
            autopos.append( nodes[-1] == (None, None) )  # If autopos is activated, need a second loop

            jp2loc.append(nass+jjj)
            loc2glob.append(nglob)
            self.g2ass.append(-1)
            self.g2sys.append(i)
            nglob += 1
            jpname.append(jpt.attrib['target'])
            jpdict[jpt.attrib['target']] = jjj

            jjj += 1

         self.autoposs.append(autopos)
         self.jpnames.append(jpname)
         self.jpdicts.append(jpdict)
         self.jp2locs.append(jp2loc)
         self.nodess.append(nodes)
         self.loc2globs.append(loc2glob)
         i += 1

      nsys = len(self.sysnames)
      connect = np.zeros((nsys,nsys)) # Connectivity matrix for systems.

      for i, (jpname, autopos, loc2globi, jp2loci, namei) in enumerate(zip(self.jpnames, self.autoposs, self.loc2globs, self.jp2locs, self.sysnames)):
         for j in range(len(jpname)):
            k = self.sysdict[jpname[j]] # Get the index of target
            connect[i,k] = 1 # Systems connectivity

            if autopos[j]: # Compute autopos stuff
               theta = math.atan2( self.ylist[k]-self.ylist[i], self.xlist[k]-self.xlist[i] )
               x = self.radius[i] * math.cos(theta)
               y = self.radius[i] * math.sin(theta)
               self.nodess[i][jp2loci[j]] = (x,y) # Now we have the position

      # Compute distances.
      connect_full = connect.copy()
      for i, hjname in enumerate(hjnames):
         for dest in hjname:
            connect_full[i, self.sysdict[dest]] = 1
      distances_strict = sp.csgraph.dijkstra(connect)
      distances_hidden = sp.csgraph.dijkstra(connect_full)

      # Use distances to compute ranged presences
      for i in range(nsys):
         for j in range(nsys):
            infos = [assets[ai] for ai in self.sysass[j]] + [vassets[ai] for ai in sysvass[j]]
            for info in infos:
               for facname, facpres in info.presences:
                  if facname not in factions:
                     continue
                  for gen in [Generator(facname, 1.)] + facinfo[facname].generators:
                     if gen.faction not in factions:
                        continue
                     fact = factions[ gen.faction ]
                     d = self.distances_hidden[i,j] if facinfo[gen.faction].useshiddenjumps else distances_strict[i,j]
                     if d <= facpres.range:
                        self.presences[i][fact] += facpres * (gen.weight / (1+d))

      #TODO distinguish between base and bonus? (Cf. same comment in src/safelanes.c)
      for presence in self.presences + [self.presass]:
         presence[:] = [max(0, p.value) for p in presence]



#!/usr/bin/env python3

from sys import stdin, stderr, argv, exit
from geometry import vec

if __name__ == '__main__':
   if '-h' in argv[1:] or '--help' in argv[1:]:
      stderr.write(
         'usage  ' + argv[0].split('/')[-1] + '\n'
         '   This is intended as a module to build graphmods,\n'
         '   i.e. programs reading a graph on stdin and outputting a modified\n'
         '   version of the graph on stdout.\n'
         '   Calling this module as an executable works as a NO-OP graphmod,\n'
         '   i.e. it outputs the same exact graph that was in input.\n'
         '   However, this has a latching effect: no output is sent until all\n'
         '   the input has been read.\n'
      )
      exit(0)

class _pos(dict):
   aux = {}
   def output( self ):
      for k, v in self.items():
         if k[0] != '_':
            v = round(v, 9)
            if k in self.aux:
               crt_aux = self.aux[k]
            else:
               crt_aux = []
            print(' '.join([k, str(v[0]), str(v[1])] + crt_aux))

class _jmp(dict):
   def output( self ):
      for i, l in self.items():
         for j, f in l.items():
            print(' '.join([i, j] + f))

   def __missing__( self, elt ):
      self[elt] = {}
      return self[elt]

class _graph():
   quiet = False

   def __init__( self ):
      self.pos=_pos()
      self.jmp=_jmp()

   def silence( self ):
      self.quiet = True

   def __del__( self ):
      if not self.quiet:
         try:
            self.pos.output()
            self.jmp.output()
            for i in stdin:
               pass
            print()
         except KeyboardInterrupt:
            exit(0)
         except BrokenPipeError:
            stderr.write('Broken output pipe. Bye !\n')
            exit(0)

graph = _graph()
ssys_pos = graph.pos
ssys_jmp = graph.jmp
no_graph_out = lambda:graph.silence()

try:
   for inp in stdin:
      if (line := inp.strip()) != '':
         l = line.split(' ')
         try:
            bname, x, y = tuple((l[:3]))
            ssys_pos[bname] = vec(x, y)
            ssys_pos.aux[bname] = l[3:]
            success = True
         except:
            success = False

         if not success:
            if len(l) >= 2:
               bname1, bname2 = tuple(l[:2])
               ssys_jmp[bname1][bname2] = l[2:]
            else:
               stderr.write('Ignored invalid line: ' + line + '\n')
               continue
      else:
         break
except KeyboardInterrupt:
   exit(0)
except BrokenPipeError:
   stderr.write('Broken input pipe. Bye !\n')
   exit(0)

if len(ssys_pos) == 0 or len(ssys_jmp) == 0:
   from traceback import extract_stack
   from os import path
   from_file = path.basename(extract_stack()[0].filename)
   stderr.write('\033[31m' + from_file + '\033[0m: empty input, bye !\n')
   exit(-1)



# The empire organigram (military part for now). Add everything you want, it may fit.

import pygraphviz as pgv

G=pgv.AGraph(directed=True)

# Ranks
G.add_subgraph(name="0")
sub = G.get_subgraph("0")
sub.graph_attr['rank']='same'
sub.add_node("His (Her?) Very Holy Majesty The Emperor Of The Universe", shape='octagon')

G.add_subgraph(name="1")
sub = G.get_subgraph("1")
sub.graph_attr['rank']='same'
sub.add_node("Holy-class", shape='octagon')

G.add_subgraph(name="2")
sub = G.get_subgraph("2")
sub.graph_attr['rank']='same'
sub.add_node("Highest-Excellency-class", shape='octagon')

G.add_subgraph(name="3")
sub = G.get_subgraph("3")
sub.graph_attr['rank']='same'
sub.add_node("Excellency-class", shape='octagon')

G.add_subgraph(name="4")
sub = G.get_subgraph("4")
sub.graph_attr['rank']='same'
sub.add_node("Eminence-class", shape='octagon')

G.add_subgraph(name="5")
sub = G.get_subgraph("5")
sub.graph_attr['rank']='same'
sub.add_node("Aristocracy", shape='octagon')

G.add_subgraph(name="6")
sub = G.get_subgraph("6")
sub.graph_attr['rank']='same'
sub.add_node("Common-class", shape='octagon')

G.add_edge("His (Her?) Very Holy Majesty The Emperor Of The Universe", "Holy-class", style='invis')
G.add_edge("Holy-class", "Highest-Excellency-class", style='invis')
G.add_edge("Highest-Excellency-class", "Excellency-class", style='invis')
G.add_edge("Excellency-class", "Eminence-class", style='invis')
G.add_edge("Eminence-class", "Aristocracy", style='invis')
G.add_edge("Aristocracy", "Common-class", style='invis')


nomNdismiss = "Nominates\n and can dismiss"

# Functions
sub = G.get_subgraph("0")
emperor = "The Emperor\n has full powers over the entire universe"
sub.add_node(emperor)

###
sub = G.get_subgraph("1")
spouse = "First Imperial Spouse"
sub.add_node(spouse)
G.add_edge( emperor, spouse, label="Play cards with" )

gal_archbishop = "Galactic High Archbishop"
sub.add_node(gal_archbishop)
G.add_edge( emperor, gal_archbishop, label="Directly obeys" )

archking = "Archking of Timbuktu\n The only dignitary allowed to pronounce death sentence\n against any holy or (highest)excellency-class dignitaries\n He only has to cut himself a finger after each execution"
sub.add_node(archking)
G.add_edge( emperor, archking )

iguana = "Pet Iguana"
sub.add_node(iguana)
G.add_edge( emperor, iguana, label="Walks" )

optician = "Optician of the Emperor\n Only person in the universe allowed to look the Emperor in the eyes\n (but his own eyes have to be burned afterwards)"
sub.add_node(optician)
G.add_edge( emperor, optician, label="Fixes his glasses" )

###
sub = G.get_subgraph("2")
concubin = "First Imperial Concubin"
sub.add_node(concubin)
G.add_edge( emperor, concubin, label="Play chess with" )
G.add_edge( spouse, concubin, label="Can dismiss\n with agreement of the Senate" )

chancelor = "High Chancellor of the Senate"
sub.add_node(chancelor)
G.add_edge( emperor, chancelor, label="Directly obeys" )
G.add_edge( iguana, chancelor, label="Can pee on" )

achbishop4 = "Fifth Galactic Archbishop"
sub.add_node(achbishop4)
G.add_edge( gal_archbishop, achbishop4, label=nomNdismiss )

kings = "Other Kings"
sub.add_node(kings)
G.add_edge( archking, kings, label="Nominates\n and can dismiss" )

cyberking = "Cyber-king of Orion\n Can oppose veto to any military action\n (except if any imperial squire opposes a veto to this veto)"
sub.add_node(cyberking)
G.add_edge( archking, cyberking, label=nomNdismiss )

iguana_spouse = "Pet Iguana's Spouse"
sub.add_node(iguana_spouse)
G.add_edge( iguana, iguana_spouse, label="Play cards with" )

butler = "Butler of the Imperial House"
sub.add_node(butler)
G.add_edge( emperor, butler, label="Directly obeys" )
G.add_edge( spouse, butler, label=nomNdismiss )

###
sub = G.get_subgraph("3")
senate = "The Senate"
sub.add_node(senate)
G.add_edge( chancelor, senate, label=nomNdismiss )

marshall = "Great Marshall of the Empire\n coordinates all large-scale military operations"
sub.add_node(marshall)
G.add_edge( butler, marshall, label=nomNdismiss )
G.add_edge( optician, marshall, label="Can dismiss" )

squire = "High Squire of the Emperor"
sub.add_node(squire)
G.add_edge( butler, squire, label=nomNdismiss )

vizir = "Great Vizir of the Empire\n Chief of the Police"
sub.add_node(vizir)
G.add_edge( butler, vizir, label=nomNdismiss )
G.add_edge( concubin, vizir, label="Can oppose a veto\n to any decision" )

###
sub = G.get_subgraph("4")
staff_senate = "Second deputee Chief of Staff of the Senate"
sub.add_node(staff_senate)
G.add_edge( chancelor, staff_senate, label=nomNdismiss )
G.add_edge( iguana_spouse, staff_senate, label="Can dismiss" )

Lsquire = "Lower Squires of the Emperor"
sub.add_node(Lsquire)
G.add_edge( squire, Lsquire, label=nomNdismiss )

###
sub = G.get_subgraph("5")
spouse_captain = "High Captain of the First Imperial Spouse's guard\n Commands all ships based on Halir\n and all the land forces on Dune"
sub.add_node(spouse_captain)
G.add_edge( spouse, spouse_captain, label=nomNdismiss )

holy_general = "Highest General of the Holy Force\n Commands all troops based in the sol system :'("
sub.add_node(holy_general)
G.add_edge( achbishop4, holy_general, label=nomNdismiss )

legate = "Legate-in-chief of the army\n Commands most Imperial land units"
sub.add_node(legate)
G.add_edge( kings, legate, label=nomNdismiss )
G.add_edge( senate, legate, label="Can dismiss\n unless any Imperial Concubin opposes a veto" )

squire_admiral = "Great Admiral of Interception\n Commands most Imperial interception forces"
sub.add_node(squire_admiral)
G.add_edge( squire, squire_admiral, label=nomNdismiss )

iguana_admiral = "Great Admiral of Bulk Forces\n Commands most Imperial heavy fleets"
sub.add_node(iguana_admiral)
G.add_edge( iguana, iguana_admiral, label=nomNdismiss )

iguana_caretaker = "Pet Iguana's Chief Caretaker"
sub.add_node(iguana_caretaker)
G.add_edge( iguana, iguana_caretaker, label="Takes care of" )
G.add_edge( senate, iguana_caretaker, label=nomNdismiss )

###
sub = G.get_subgraph("6")
concubin_captain = "Low Captain of the First Imperial Concubin's guard\n Commands all the land forces based on Halir"
sub.add_node(concubin_captain)
G.add_edge( concubin, concubin_captain, label=nomNdismiss )

senate_admiral = "Great Admiral of the Senate's Fleet\n Commands the senate's fleet, based in Arcturus Gamma"
sub.add_node(senate_admiral)
G.add_edge( staff_senate, senate_admiral, label=nomNdismiss )
G.add_edge( senate, senate_admiral, label="Controls his actions" )

genetician = "Genetician of the Archking"
sub.add_node(genetician)
G.add_edge( archking, genetician, label="Ensures he always has\n enough fingers" )
G.add_edge( gal_archbishop, genetician, label="Can Dismiss" )

interpreter = "Pet Iguana's Spouse's Interpreter"
sub.add_node(interpreter)
G.add_edge( iguana_spouse, interpreter, label="Interprets" )

deputee_caretaker = "Second Pet Iguana's Deputee Caretaker"
sub.add_node(deputee_caretaker)
G.add_edge( iguana_caretaker, deputee_caretaker, label=nomNdismiss )

optician_o = "Optician of the Emperor's optician."
sub.add_node(optician_o)
G.add_edge( optician, optician_o, label="Ensures he can serve\n more than one time" )

bodyguard = "Imperial bodyguards\n Can serve as a special commando"
sub.add_node(bodyguard)
G.add_edge( butler, bodyguard, label="Directly obeys" )

squire_squadron = "Squadron of the Imperial squires\n Elite fighter squadron."
sub.add_node(squire_squadron)
G.add_edge( Lsquire, squire_squadron, label="Directly obeys" )

militia = "Great Master of the militias\n coordinates the 362 irregular militias\n that have been attached to the army"
sub.add_node(militia)
G.add_edge( vizir, militia, label=nomNdismiss )

###
iguana_captain = "Low Captain of the Pet Iguana's guard\n Commands all ships based on planet Oma\n and all land troops on Dolmen"
G.add_node(iguana_captain)
G.add_edge( deputee_caretaker, iguana_captain, label=nomNdismiss )

# Draw
G.layout(prog='dot')
G.draw("organigram.pdf")



#!/usr/bin/env python3

import xml.etree.ElementTree as ET
import argparse

parser = argparse.ArgumentParser( description='Script to generate naevpedia markdown from ship XML files.')
parser.add_argument('path', metavar='PATH', type=str, help='Name of the ship XML file.')
parser.add_argument('-o', type=str, help='Output path.' )

args, unknown = parser.parse_known_args()

def tags_recursive( root, d, curpath ):
   for tag in root:
      dataname = tag.tag if curpath=='' else curpath+'/'+tag.tag
      # has children
      if len(tag):
         tags_recursive( tag, d, dataname )
      # no children
      else:
         d[dataname] = tag.text

tree = ET.parse( args.path )
root = tree.getroot()
name = root.get('name')
d = {'name':name, 'general/description':""}
tags_recursive( root, d, '' )
#print( d )

outstr = f"""---
title: "{d['name']}"
cond: "return outfit.get([[{d['name']}]]):known()"
---
<% o = outfit.get([[{d['name']}]]) %>
"""
# We don't want any substitution below if possible
outstr += """
<%
local lg = require "love.graphics"
local luatk = require "luatk"
function mainimg ( mw )
   return luatk.newImage( nil, -10, 0, 256, 256, lg.newImage(o:icon()), {
      frame = {0.05,0.05,0.05},
      bkg = {0,0,0},
   } )
end

function wgtdescription( mw, tw )
   return luatk.newText( nil, 10, 0, tw, nil, o:description( nil ) )
end

function wgtsummary( mw, tw )
   return luatk.newText( nil, 10, 0, tw, nil, o:summary( nil, true ) )
end
%>
<widget mainimg />

# <%= o:name() %>

<widget wgtdescription />

## Properties

<widget wgtsummary />

## Availability

<%
   local availability = {}
   for i,sys in ipairs(system.getAll()) do
      for j,spb in ipairs(sys:spobs()) do
         if spb:known() and inlist( spb:outfitsSold(), o ) then
            table.insert( availability, sys )
            break
         end
      end
   end
   table.sort( availability, function ( a, b )
      return a:jumpDist() < b:jumpDist()
   end )
%>
<% if #availability > 0 then %>
Places where <%= o:name() %> are sold are shown in #ggreen#0.
<% function map ( mw, tw )
   local luatk_map = require "luatk.map"
   local w = math.min( mw-200, tw-20 )
   local m = luatk_map.newMap( nil, 10, 0, w, w * 9 / 16, {
      binaryhighlight = function ( sys )
         if inlist( availability, sys ) then
            return true
         end
         return false
      end
   } )
   if #availability > 0 then
      m:center( availability[1]:pos(), true )
   end
   return m
end %>
<widget map />
<% else %>
You are not aware of anywhere where <%= o:name() %> are available for purchase.
<% end %>
"""

with open(args.o, 'w', encoding='utf-8', newline='\n') as f:
   f.write( outstr.replace('\r\n','\n') )



#!/usr/bin/env python3


if __name__ != '__main__':
   raise Exception('This module is only intended to be used as main.')

from sys import stderr, argv, exit

if JUST_LIST := '-L' in argv[1:]:
   argv.remove('-L')

if argv[1:]:
   stderr.write(
      'usage: ' + argv[0].split('/')[-1] + ' [-L]\n'
      '  Reads a graph file on stdin, outputs a graph on stdout.\n'
      '  Intended as a postprocessing for neato output.\n'
      '  If -L is set, only outputs the list of sys it would modify.\n'
   )
   exit(0)

from geometry import symmetry
from graphmod import ssys_pos, ssys_jmp, no_graph_out
from smoothen import smoothen, smoothen_induced, circleify
from graph_vaux import ssys_others

L = {'eiderdown', 'gilligans_tomb', 'adraia', 'vanir', 'botarn', 'monogram', 'kraft', 'pike'}
sirius_circle = {'palovi',} | L
tradelane = { k: {s for (s, t) in ssys_jmp[k].items() if 'tradelane' in t and not {k, s} < L} for k in ssys_pos }

rem_tl = lambda l, s: {k: {} if k in s else (v - s) for k, v in l.items()}
tradelane = rem_tl(tradelane, {'point_zero'})

stellars = {k for k in ssys_pos if 'stellarwind' in ssys_others(ssys_pos, k)} - {'c59'}
abh_circle = {'ngc11935', 'ngc5483', 'ngc7078', 'ngc7533', 'octavian', 'copernicus', 'ngc13674', 'ngc1562', 'ngc2601'}

trad_l_pos = smoothen(ssys_pos, tradelane)
stel_w_pos = smoothen_induced(
   ssys_pos | {'pilatis': symmetry(ssys_pos['defa'])(ssys_pos['oberon'])},
   ssys_jmp, stellars, hard= True)

if JUST_LIST:
   no_graph_out()
   print('\n'.join(trad_l_pos.keys() | stel_w_pos.keys() | sirius_circle | abh_circle))
else:
   ssys_pos |= circleify(ssys_pos, abh_circle, 'anubis_black_hole')
   ssys_pos |= circleify(ssys_pos, sirius_circle, 'suna', hard= True)
   ssys_pos |= stel_w_pos | trad_l_pos



#!/usr/bin/env python3
# This file generates safe lanes

# TODO the presence calculation is not the same as the current Naev
# implementation!! Faction generators and virtual assets are not taken into
# account.

import copy
import math
import numpy as np
import scipy.sparse as sp
from sksparse.cholmod import cholesky

from lanes_perf import timed, timer
from lanes_ui import printLanes
from lanes_systems import Systems
from union_find import UnionFind

JUMP_CONDUCTIVITY = .001  # TODO : better value
DISCONNECTED_THRESHOLD = 1e-12
FACTIONS_LANES_BUILT_PER_ITERATION = 1
MAX_ITERATIONS = 20
MIN_ANGLE = 10*math.pi/180 # Path triangles can't be more acute.
ALPHA = 9  # Lane efficiency parameter.
LAMBDA = 2e10 # Regularization term.

def inSysStiff( nodess, factass, g2ass, loc2globNs ):
   '''Compute insystem paths. TODO maybe : use Delauney triangulation instead?'''
   si  = []  # Element to build the sparse default matrix
   sj  = []
   sv  = []
   sil = [] # Lanes in local numerotation
   sjl = []
   sr  = [] # Lanes reserved for a faction
   loc2globs = []
   system = [] # Tells which system the lane is in

   crit = math.cos(MIN_ANGLE)
   i  = 0
   ii = 0

   for k, nodes in enumerate(nodess):
      loc2glob = []
      loc2globN = loc2globNs[k]
      for n in range(len(nodes)):
         xn, yn = nodes[n]

         na = g2ass[loc2globN[n]]  # Find global asset numerotation of local node
         if na>=0:
            fn = factass[na]
         else: # It's not an asset
            fn = -1

         for m in range(n): # Only m<n, because symmetry
            xm, ym = nodes[m]
            ma = g2ass[loc2globN[m]]
            if ma>=0:
               fm = factass[ma]
            else: # It's not an asset
               fm = -1

            lmn = math.hypot( xn-xm, yn-ym )

            # Check if very close path exist already.
            forbidden = False
            for p in range(len(nodes)):
               if n==p or m==p:
                  continue
               xi, yi = nodes[p]

               # The scalar product should not be too close to 1
               # That would mean we've got a flat triangle
               lni = math.hypot(xn-xi, yn-yi)
               lmi = math.hypot(xm-xi, ym-yi)
               dpn = ((xn-xm)*(xn-xi) + (yn-ym)*(yn-yi)) / ( lmn * lni )
               dpm = ((xm-xn)*(xm-xi) + (ym-yn)*(ym-yi)) / ( lmn * lmi )
               if (dpn > crit and lni < lmn) or (dpm > crit and lmi < lmn):
                  forbidden = True
                  break

            if forbidden:
               continue

            si.append( i+n )
            sj.append( i+m )
            sv.append( 1/lmn )
            sil.append( n )
            sjl.append( m )
            sr.append((fn,fm)) # This tells who is allowed to build along the lane
            system.append(k)

            loc2glob.append(ii)
            ii += 1

      i += len(nodes)
      loc2globs.append(loc2glob)

   return ( si, sj, sv, sil, sjl, loc2globs, sr, system )


class SafeLaneProblem:
   '''Representation of the optimization problem to be solved.'''
   def __init__( self, systems ):
      si0 = [] #
      sj0 = [] #
      sv0 = [] # For the construction of the sparse weighted connectivity matrix
      biconnect = UnionFind(len(systems.sysnames))  # Partitioning of the systems by reversible-jump connectedness.

      for i, (jpname, loc2globi, jp2loci, namei) in enumerate(zip(systems.jpnames, systems.loc2globs, systems.jp2locs, systems.sysnames)):
         for j in range(len(jpname)):
            k = systems.sysdict[jpname[j]] # Get the index of target

            jpnamek = systems.jpdicts[k]
            loc2globk = systems.loc2globs[k]
            jp2lock = systems.jp2locs[k]

            if i < k or namei not in jpnamek:
               continue # We only want to consider reversible jumps, and only once per pair.

            m = jpnamek[namei] # Index of system i in system k numerotation

            si0.append(loc2globi[jp2loci[j]]) # Assets connectivity (jp only)
            sj0.append(loc2globk[jp2lock[m]]) #
            sv0.append(JUMP_CONDUCTIVITY)
            biconnect.union(i, k)

      # Create anchors to prevent the matrix from being singular
      # Anchors are jumpoints, there is 1 per connected set of systems
      # A Robin condition will be added to these points and we will check the rhs
      # thanks to them because in u (not utilde) the flux on these anchors should
      # be 0, otherwise, it means that the 2 non-null terms on the rhs are on
      # separate sets of systems.
      self.anchors = [systems.loc2globs[i][0] for i in biconnect.findall() if systems.loc2globs[i]]

      # Get the stiffness inside systems
      self.default_lanes = (si0,sj0,sv0)
      self.internal_lanes = inSysStiff( systems.nodess, systems.factass, systems.g2ass, systems.loc2globs )
      self.vertex_factions = [[] for _ in systems.g2ass] # Factions that have visited each vertex (in the global enumeration)


def buildStiffness( problem, activated, systems ):
   '''Computes the conductivity matrix'''
   def_si, def_sj, def_sv = problem.default_lanes[:3]
   int_si, int_sj, int_sv0  = problem.internal_lanes[:3]

   # Take activated lanes into account
   int_sv = int_sv0.copy()
   for i in range(len(int_sv0)):
      if activated[i]:
         int_sv[i] = (ALPHA+1)*int_sv0[i]

   # Assemble both lists
   si = def_si + int_si
   sj = def_sj + int_sj
   sv = def_sv + int_sv

   # Build the sparse matrix
   sii = si + sj + si + sj
   sjj = si + sj + sj + si
   svv = sv*2 + [-x for x in sv]*2

   mu = max(sv)  # Just a parameter to make a Robin condition that does not spoil the spectrum of the matrix
   for i in problem.anchors:
      sii.append(i)
      sjj.append(i)
      svv.append(mu)

   stiff = sp.csc_matrix( ( svv, (sii, sjj) ) )
   # Rem : it may become mandatory at some point to impose nb of dofs

   return stiff

@timed
def compute_PPts_QtQ( problem, utilde, systems ):
   '''Gives the matrix that computes penibility form potentials.
      By chance, this does not depend on the presence of lane.'''
   nfact = len(systems.presences[0])
   nass = len(systems.ass2g)

   # First : compute the (sparse) matrix that transforms utilde into u
   pi = []
   pj = []
   pv = []

   di = [[] for i in range(nfact)]
   dv = [[] for i in range(nfact)]

   for i, ai in enumerate(systems.ass2g):
      facti = systems.factass[i]
      presi = systems.presass[i]
      for j in range(i):# Stuff is symmetric, we use that
         if utilde[ai,j] <= DISCONNECTED_THRESHOLD:
            continue
         ij = (i*(i-1))//2 + j # Multiindex

         # Just a substraction
         pi.append(i)
         pj.append(ij)
         pv.append(1)
         pi.append(j)
         pj.append(ij)
         pv.append(-1)

         factj = systems.factass[j]
         presj = systems.presass[j]

         # Build the diagonal weights
         if facti >= 0:
            di[facti].append(ij)
            dv[facti].append(presi)
         if factj >= 0:
            di[factj].append(ij)
            dv[factj].append(presj)

   P = sp.csr_matrix( ( pv, (pi, pj) ) )
   # P*utilde^T = u^T

   Pp = ( P @ sp.csr_matrix( (dv[k], (di[k], di[k])), (P.shape[1], P.shape[1]) ) for k in range(nfact) )
   PPl = [ (Ppk @ Ppk.transpose()).todense() for Ppk in Pp ]

   si, sj = problem.internal_lanes[:2]

   # Then : the matrix that gives penibility on each internal lane from u
   qi = []
   qj = []
   qv = []
   for i in range(len(si)):
      # Just a substraction (again)
      qi.append(i)
      qj.append(si[i]) # TODO : check the transpose
      qv.append(1)
      qi.append(i)
      qj.append(sj[i])
      qv.append(-1)

   Q = sp.csr_matrix( (qv, (qi, qj)), (len(si), len(systems.g2ass)) )
   # Q*u = p

   return PPl, Q.transpose() @ Q


@timed
def getGradient( problem, u, lamt, PPl, pres_0, activated, systems, iters_done ):
   '''Get the gradient from state and adjoint state.
      Luckily, this does not depend on the stiffness itself (by linearity wrt. stiffness).'''
   si, sj, sv = problem.internal_lanes[:3]
   sr = problem.internal_lanes[6] # Tells who has right to build on each lane
   sy = problem.internal_lanes[7] # Tells the system
   vf = problem.vertex_factions

   sz = len(si)

   nfact = len(PPl)
   gl = []
   for k in range(nfact): # .2
      # Build the list of all interesting dofs
      myDofs = []
      for i in range(len(systems.loc2globs)):
         if pres_0[i][k] <= 0: # Does this faction have presence here ?
            continue

         loc2glob = systems.loc2globs[i] # Idices of the assets in the global numerotation
         myDofs = myDofs + loc2glob


      lal = np.zeros(lamt.shape)
      lal[myDofs,:] = lamt[myDofs,:] @ PPl[k]
      #lal = lamt @ PPl[k]
      glk = np.zeros((sz,1))

      for i in range(sz):

         if activated[i]: # The lane has already been activated: no need to re-compute its gradient
            continue

         if pres_0[sy[i]][k] <= 0: # Does this faction have presence here ?
            continue

         sis = si[i]
         sjs = sj[i]

         faction_may_build = ((k in sr[i]) or (sr[i] == (-1, -1))) and ((k in vf[sis] or k in vf[sjs]) or not iters_done)
         if not faction_may_build:
            continue

         LUTll = np.dot( lal[[sis,sjs],:] , u[[sis,sjs],:].T )
         glk[i] = ALPHA*sv[i] * ( LUTll[0,0] + LUTll[1,1] - LUTll[0,1] - LUTll[1,0] )

      gl.append(glk)
   return gl


def activateBestFact( problem, gl, activated, Lfaction, pres_c, pres_0, iters_done ):
   '''Activates the best lane in each system for each faction. Returns the number activated. '''
   # Find lanes to keep in each system
   si, sj, sv = problem.internal_lanes[:3]
   lanesLoc2globs, sr = problem.internal_lanes[5:7]
   vf = problem.vertex_factions
   nsys = len(lanesLoc2globs)
   nactivated = 0
   nfact = len(pres_c[0])

   g1l = [gl[k] * np.c_[sv] for k in range(nfact)]

   for i in range(nsys):
      lanesLoc2glob = lanesLoc2globs[i]

      if all(activated[k] for k in lanesLoc2glob):
         continue

      # Faction with more presence choose first
      ploc = pres_0[i]
      sind = np.argsort(ploc)
      sind = np.flip(sind)

      for ff in range(nfact):
         f = sind[ff]

         for lane in range(FACTIONS_LANES_BUILT_PER_ITERATION):
            if pres_c[i][f] <= 0.: # This faction has no presence
               continue

            gloc = g1l[f][lanesLoc2glob]
            ind1 = np.argsort(gloc.transpose()) # For some reason, it does not work without transpose :/
            ind = [lanesLoc2glob[k] for k in ind1[0]]

            # Find a lane to activate
            prev_nactivated = nactivated
            for k in ind:
               faction_may_build = ((f in sr[k]) or (sr[k] == (-1, -1))) and ((f in vf[si[k]] or f in vf[sj[k]]) or not iters_done)
               cost = 1/sv[k] / systems.dist_per_presence[f] + systems.lane_base_cost[f]
               score = g1l[f][k, 0] + LAMBDA
               if (not activated[k]) and pres_c[i][f] >= cost and faction_may_build and score < 0:
                  pres_c[i][f] -= 1/sv[k] / systems.dist_per_presence[f]
                  for vi in si[k], sj[k]:
                     if f not in vf[vi]:
                        vf[vi].append(f)
                  activated[k] = True
                  Lfaction[k] = f
                  nactivated += 1
                  break

            # The faction did not activate anything:
            # There is no hope in activating anything anymore.
            if nactivated == prev_nactivated:
               pres_c[i][f] = -1

   return nactivated


@timed
def optimizeLanes( systems, problem ):
   sz = len(problem.internal_lanes[0])
   activated = [False] * sz # Initialization : no lane is activated
   Lfaction = [-1] * sz;
   pres_c = copy.deepcopy(systems.presences)
   nfact = len(systems.presences[0])

   # TODO : It could be interesting to use sparse format for the RHS as well (see)
   ftilde = np.eye( len(systems.g2ass) )
   ftilde = ftilde[:,systems.ass2g] # Keep only lines corresponding to assets

   for i in range(MAX_ITERATIONS):
      stiff = buildStiffness( problem, activated, systems ) # ~0.008 s
      stiff_c = cholesky(stiff) #.001s

      # Compute direct and adjoint state
      utilde = stiff_c.solve_A( ftilde ) # .004 s

      # Compute QQ and PP, and use utilde to detect connected parts of the mesh
      # It's in the loop, but only happens once
      if i == 0: # .5 s
         PPl, QQ = compute_PPts_QtQ( problem, utilde, systems )

      rhs = - QQ @ utilde
      lamt = stiff_c.solve_A( rhs ) #.010 s

      # Compute the gradient and activate. Runtime depends on how many degrees of freedom remain.
      gl = getGradient( problem, utilde, lamt, PPl, pres_c, activated, systems, i )

      if not activateBestFact( problem, gl, activated, Lfaction, pres_c, systems.presences, i ):
         break

   print('Converged in', i+1, 'iterations.')
   return (activated, Lfaction)

if __name__ == "__main__":
   systems = Systems()
   problem = SafeLaneProblem( systems )

   activated, Lfaction = optimizeLanes( systems, problem )
   printLanes( problem, activated, Lfaction, systems )



#!/usr/bin/env python3

if __name__ != '__main__':
   raise Exception('This module is only intended to be used as main.')


from sys import argv, stderr
from graph_vaux import color_values, ssys_color, ssys_nebula, ssys_others


def main( colour = False, fixed_pos = False ):
   from graphmod import ssys_pos as pos, ssys_jmp as E, no_graph_out
   no_graph_out()

   if colour:
      colours = { k: color_values[ssys_color(pos, k)] for k in pos }
      nebula = { k: ssys_nebula(pos, k) for k in pos if ssys_nebula(pos, k) is not None }
      others = { k: ssys_others(pos, k) for k in pos }
      V = {k:' '.join(l[1:]) for k, l in pos.aux.items()}
   else:
      V = {k:' '.join(l) for k, l in pos.aux.items()}

   print(
      'graph g{\n'
      '\tepsilon=0.000001\n'
      '\tmaxiter=2000\n'
   )

   # 1inch=72pt
   if fixed_pos:
      print('\tgraph [overlap=true]')
      factor = 0.7
   else:
      print('\tgraph [overlap=false]')  #'\toverlap=voronoi'
      factor = 0.7

   reflen = 0.5
   print(
      '\tinputscale=72\n'
      '\tnotranslate=true\n' # don't make upper left at 0,0
      '\tnode[fixedsize=true,shape=circle,penwidth=0,color=white,fillcolor=grey,style="filled"]\n'
      '\tnode[width=0.5]\n'
      '\tedge[len=' + str(reflen) + ']'
   )
   if fixed_pos:
      print('\tnode[pin=true]')

   for i in V:
      if i[0] == '_' and fixed_pos:
         continue
      # Don't include disconnected systems
      if E[i] != {} or fixed_pos:
         s = '\t"' + i + '" ['
         if i[0] != '_':
            (x, y) = round(pos[i] * factor, 9)
            fixed = '!' if fixed_pos else ''
            s += 'pos="' + str(x) + ',' + str(y) + fixed + '";'
         label = V[i]
         for t in {'-': '- ', ' ': '\\n', 'Test\\nof': 'Test of'}.items():
            label = label.replace(*t)
         s += 'label="' + label + '"'

         if colour:
            if i in nebula:
               for lev, col in {
                  0:'blue',
                  10:'purple',
                  45:'fuchsia',
                  75:'deeppink',
                  100:'red'
               }.items():
                  if nebula[i] <= lev:
                     s += ';fontcolor=' + col
                     break
            c = {
               'stellarwind': 'lightskyblue',
               'haze':        'pink',
               'plasmastorm': '".833 .2 1"'
            }
            for o in others[i]:
               if o in c:
                  s += 'penwidth=4.0;color=' + c[o]
                  break
            cols = [int(255.0*(f/3.0+2.0/3.0)) for f in colours[i]]
            rgb = ''.join([('0'+(hex(v)[2:]))[-2:] for v in cols])
            s += ';fillcolor="#' + rgb + '"'

         if i == 'sol':
            s += ';color=red'

         print(s + ']')

   # all others are virtual
   print('\tnode [label="",style=invis]')

   def jump_c_f(aux):
      jmp_c = {'hidden': 'purple', 'new': 'green', 'fake': 'red'}
      srcc = [jmp_c[a] for a in aux if a in jmp_c]
      if {'purple', 'green'} <= set(srcc):
         return 'gold'
      else:
         return (srcc + ['black']) [0]

   for i in V:
      for dst, aux in E[i].items():
         suff = []
         if 'virtual' in aux:
            continue;
         elif 'tradelane' in aux:
            suff += ['style=bold', 'penwidth=4.0']
         elif 'fake' in aux:
            suff += ['weight=0']

         srcc = jump_c_f(aux)
         if (oneway := i not in E[dst]) or i < dst:
            dstc = 'grey' if oneway else jump_c_f(E[dst][i])

            if srcc != dstc:
               suff += ['color="' + srcc + ';0.5:' + dstc + '"']
            elif srcc!= 'black':
               suff += ['color="' + srcc + '"']
            suff = '[' + ';'.join(suff) + ']' if suff else ''
            print('"'.join(['\t', i, '--', dst, suff]))

   print(
      '\tedge[len=' + str(reflen) + ']\n'
      '\tedge[style="dashed";color="grey";penwidth=1.5]'
   )
   for f, k in E.items():
      for t, aux in k.items():
         if fixed_pos or 'virtual' not in aux:
            continue
         try:
            l = float(aux[0])
         except:
            l = 1.0

         prop = []
         if f[0] == '_' or t[0] == '_':
            prop += ['style="invis"']

         if l != 1.0:
            prop += ['len='+str(l*reflen)]

         prop = ';'.join(prop)
         if prop != '':
            prop = ' [' + prop + ']'
         print('\t"' + f + '"--"' + t + '"' + prop)
   print('}')

if colour := '-c' in argv:
   argv.remove('-c')

if keep := '-k' in argv:
   argv.remove('-k')

if '-h' in argv[1:] or '--help' in argv[1:]:
   print(
      'usage: ', argv[0], '[-c]', '[-k]', '\n'
      '  Outputs the graph in dot format.\n'
      '  By default, interprets the vertex aux as the name.\n'
      '  If -c is set, interprets the first aux field as colour.\n'
      '  If -k is set, the nodes have the keep_position marker.\n'
      'Examples:\n'
      '  > ./utils/ssys/ssys2dot.py -k | neato -Tpng > before.png\n'
      '  > ./utils/ssys/ssys2dot.py | neato -Tpng > after.png\n'
      '  > display before.png after.png'
   )
else:
   if argv[1:]:
      stderr.write('Ignored: "' + '", "'.join(argv[1:]) + '"\n')

   main(colour, keep)



#!/usr/bin/env python3

from subprocess import Popen, PIPE
from os import getpid

class sedserver:
   def __init__( self, cmd = [__file__.replace('.py', '.sed'), '-u'] ):
      self.proc = Popen(cmd, stdin=PIPE, stdout=PIPE)
      self.parent = getpid()

   def __del__( self ):
      self.proc.terminate()

   def __call__( self, inp ):
      line = inp.rstrip('\n') + '\n'
      line = line.encode('ISO-8859-1').lower()
      self.proc.stdin.write(line)
      self.proc.stdin.flush()
      return self.proc.stdout.readline().decode('ISO-8859-1').rstrip('\n')

   def mine( self ):
      return self.parent == getpid()

def end_xml_name():
   global xml_filter
   xml_filter = None

end_xml_name()

def xml_name( s ):
   global xml_filter
   if xml_filter is None or not xml_filter.mine():
      xml_filter = sedserver()
   return xml_filter(s)

if __name__ == '__main__':
   from sys import stdin
   for i in stdin:
      print(xml_name(i.strip()))



#!/usr/bin/env python3

import sys
from os import path
dat_dir = path.realpath(path.join(path.dirname(__file__), '..', '..', '..'))
script_dir = path.realpath(path.join(dat_dir, '..', 'utils'))
sys.path.append(script_dir)
sys.path.append(path.realpath(path.join(script_dir, 'outfits')))
from outfit import outfit, naev_xml
from getconst import PHYSICS_SPEED_DAMP

# Global slack to make it easier to pull off by increasing the time given
SLACK = 1.1

if __name__ != '__main__':
   raise Exception('This is intended to be used as main')

from sys import argv

if len(argv) == 1:
   argv += ['-']

if argv[1] == '-q':
   argv.remove('-q')
   def info(*t):
      pass
else:
   info = lambda *t: sys.stderr.write(' '.join(t) + '\n')

# These are computed by the game and displayed in the race menu
names = ['Peninsula', 'Smiling Man', 'Qex Tour']
length = [23745, 40554, 36886]
#the lengths are overestimated by approx 15%
length = [n / 1.15 for n in length]

configs = [
   {
      'ship': 'neutral/hyena', 'engine': 'small/nexus_dart_160_engine',
      'AB': 'unicorp_light_afterburner',
   },{
      'ship': 'neutral/hyena', 'engine': 'small/nexus_dart_160_engine',
      'AB': 'unicorp_light_afterburner',
   },{
      'ship': 'neutral/hyena', 'engine': 'small/nexus_dart_160_engine',
      'AB': 'unicorp_light_afterburner',
   },{
      'ship': 'soromid/soromid_reaver', 'engines': 'small/tricon_zephyr_engine',
      'accessory': 'racing_trophy_bronze',
      'AG': 'adrenal_gland_ii',
      'skill': 'wanderer'
   },{
      'ship': 'soromid/soromid_reaver', 'engines': 'small/tricon_zephyr_engine',
      'accessory': 'racing_trophy_bronze',
      'AG': 'adrenal_gland_ii',
      'skill': 'wanderer'
   },{
      'ship': 'soromid/soromid_reaver', 'engines': 'small/tricon_zephyr_engine',
      'accessory': 'racing_trophy_bronze',
      'AG': 'adrenal_gland_ii',
      'skill': 'wanderer'
   },{
      'ship': 'pirate/pirate_revenant', 'engine': 'medium/tricon_cyclone_engine',
      'accessory': 'racing_trophy_silver',
      'structure': ('compact_lightsail', 2),
      'AG': 'adrenal_gland_iii',
      'skill': 'wanderer'
   },{
      'ship': 'pirate/pirate_revenant', 'engine': 'medium/tricon_cyclone_engine',
      'accessory': 'racing_trophy_silver',
      'structure': ('compact_lightsail', 2),
      'AG': 'adrenal_gland_iii',
      'skill': 'wanderer'
   },{
      'ship': 'pirate/pirate_revenant', 'engine': 'medium/tricon_cyclone_engine',
      'accessory': 'racing_trophy_silver',
      'structure': ('compact_lightsail', 2),
      'AG': 'adrenal_gland_iii',
      'skill': 'wanderer'
   }
]

ab_ratio = [0.05, 0.198, 0.402, 0.595, 0.602, 0.69, 0.885, 0.89, 0.91]

where = {
   'ship': 'ships/',
   'engine': 'outfits/core_engine/',
   'engines': 'outfits/core_engine/',
   'accessory': 'outfits/accessory/',
   'structure': 'outfits/structure/',
   'AB': 'outfits/utility',
   'AG': 'outfits/bioship/skills/',
   'skill': 'outfits/bioship/skills/',
}


def config_to_str(c):
   s = ''
   for k, t in c.items():
      if not isinstance(t, tuple):
         t = (t, 1 if k != 'engines' else 2)
      s += t[0].split('/')[-1]
      if t[1] != 1:
         s += ' x' + str(t[1])
      s += ' '
   return s

def fmt_tim(f):
   if m := int(f//60):
      return str(m) + "'" + str(round(f-m*60, 1))
   else:
      return str(round(f, 1))

race = [names[i] + ' ' + j for j in ['Bronze', 'Silver', 'Gold'] for i in range(3)]
out={}
for nj, j in enumerate(['Bronze', 'Silver', 'Gold']):
   for ni, i in enumerate(names):
      ra = j + ' ' + i
      l = length[ni]
      ij = nj*3 + ni
      c = configs[ij]
      ab = ab_ratio[ij]
      total = {
         'speed': 0, 'accel': 0,
         'speed_fact': 1.0, 'accel_fact': 1.0,
         'speed_AB': 1.0, 'accel_AB': 1.0, 'time_speedup_AB': 1.0}
      for k, v in c.items():
         if isinstance(v,tuple):
            v, count = v
         else:
            count = 1
         fnam = path.join(dat_dir, where[k], v + '.xml')
         if k == 'engine':
            o = outfit(fnam)
            o.stack()
         elif k == 'engines':
            o = outfit(fnam)
            o.stack(o)
         else:
            o = naev_xml(fnam)
         for f in ['speed', 'speed_mod', 'accel', 'accel_mod', 'time_speedup']:
            if r := o.find(f):
               r *= count
               if k in {'AB', 'AG'}:
                  total[f + '_AB'] *= 1.0 + (r/100.0)
               elif (w:= f.find('_mod')) != -1:
                  total[f[:w] + '_fact'] *= 1.0 + (r/100.0)
               else:
                  total[f] += r
      for f in ['speed', 'accel']:
         total[f] *= total[f + '_fact']
         del total[f + '_fact']
      in_AB = {k:total[k]*total[k + '_AB'] for k in ['speed', 'accel']}
      final = total['speed'] + total['accel']/PHYSICS_SPEED_DAMP
      final_AB = (in_AB['speed'] + in_AB['accel']/PHYSICS_SPEED_DAMP) * total['time_speedup_AB']

      vals = [(a[0].split('/')[-1], a[1]) if isinstance(a, tuple) else a.split('/')[-1] for a in c.values()]
      info('\n' + config_to_str(c))
      s = 'normal: ' + str(round(final))
      s += '  ' + 'AB (\033[36m' + str(round(100.0*ab)) + '%\033[m' + ' of track) ' + str(round(final_AB))
      final = ab*final_AB + (1 - ab)*final
      info(s, '->', '\033[36m' + str(round(final)) + '\033[m')
      info(ra + ':','\033[36;1m' + fmt_tim(l/final) + '\033[m')
      out[i] = out.get(i, {})
      out[i][j] = round(SLACK * l/final, 1)
info()

with open(argv[1], 'wt') if argv[1] != '-' else sys.stdout as fp:
   fp.write('return {\n')
   for track, times in out.items():
      fp.write(' '*3 + '[' + repr(track) + ']= ')
      fp.write(str(times).replace(':', '=').replace("'",'') + ',\n')
   fp.write('}\n')



import sys
from os import path
script_dir = path.join(path.dirname(__file__), '..', '..', '..', 'utils')
sys.path.append(path.realpath(script_dir))
sys.path.append(path.realpath(path.join(script_dir, 'outfits')))
from outfit import outfit

INPUT = sys.argv[1]
OUTPUT = sys.argv[2]

def read():
   o = outfit(INPUT)
   o.save_as(OUTPUT)
   if o.is_multi:
      o['outfit']['specific']['multicore_args'] = ["require('outfits.lib.set').set"]
   return o



#!/usr/bin/env python3

# SPDX-Licence-Identifier: GPL-3.0-or-later

# TODO: antialias edges of model
# TODO: diffuse
# TODO: file format: interleaved, extendable
# TODO: uniform buffer object
# TODO: https://stackoverflow.com/questions/50806126/why-are-textures-displayed-incorrectly-when-using-indexed-rendering-gldraweleme
# TODO: handle material switches within one object (combine into atlas texture? texture array? multiple draws?)
# TODO: bump map
# TODO: specral, emit, l (used in peacemaker.obj?)
# https://people.cs.clemson.edu/~dhouse/courses/405/docs/brief-mtl-file-format.html
# illum 2 is Blinn–Phong reflection model
# map files in a .mtl are square power of two
# TODO: make sure bump mapping is correct

import math
import argparse

from OpenGL.GL import *

from obj_view.parse import parse_obj
from obj_view.window import Window
from obj_view.fb import Framebuffer
from obj_view.render import ObjProgram, RenderObject


parser = argparse.ArgumentParser()
parser.add_argument('obj')
parser.add_argument('--rot', type=int, default=0)
parser.add_argument('--res', type=int, default=256)
parser.add_argument('--save')
parser.add_argument('--exit', action='store_true')
args = parser.parse_args()

rot = args.rot * math.pi / 180

window = Window(rot)
ship = parse_obj(args.obj)

if args.save is not None:
   fb = Framebuffer(args.res, args.res)

   glEnable( GL_FRAMEBUFFER_SRGB )

   glViewport(0, 0, args.res, args.res)

   program = ObjProgram()
   program.use()

   glClearColor(0., 0., 0., 0.)
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

   for obj in ship.values():
      obj = RenderObject(program, obj)
      program.draw(obj, args.res, args.res, rot)

   fb.save(args.save)
elif not args.exit:
   window.main_loop(ship)



#!/usr/bin/env python3
import helper as h
N_ = lambda text: text
data = h.read()

data['outfit']['@name'] = N_('Neutralizer')

general = data['outfit']['general']
general['rarity'] = 2
general['description'] = N_("Based on the Heavy Ion Cannon, the Neutralizer is proof to the importance of specialized engineering. With most of the hardware and software significantly modified or replaced, the Neutralizer shows significant improvements in nearly all facets. Due to the fact that few were made, it is a fairly rare find, but an import asset in a pilot's arsenal.")
general['$price'] += 150e3
general['$mass'] += -2
general['$cpu'] = round(general['$cpu'] * 0.92)

specific = data['outfit']['specific']
specific['$range'] *= 1.1
specific['$falloff'] *= 1.2
specific['$delay'] *= 1.05
specific['$energy'] *= 1.05
specific['$swivel'] *= 1.5
specific['$trackmax'] = round(specific['$trackmax'] * 0.8333333333)
specific['lua'] = 'outfits/lib/matrix_sell.lua'

damage = specific['damage']
damage['$disable'] *= 1.2
damage['$physical'] *= 1.25
damage['$penetrate'] = round(damage['$penetrate'] * 1.15)

data.save()



#!/usr/bin/env python3

if __name__ != '__main__':
   raise Exception('This module is only intended to be used as main.')


from sys import argv, stderr
from graph_vaux import color_values

if argv[1:]:
   stderr.write(
      'usage: ' + argv[0].split('/')[-1] + '\n'
      '  Reads a graph file on stdin, outputs a graph on stdout.\n'
      '  Adds a few virtual edges.\n'
   )
   exit(0)

# In the form: (from, to [, length])
virtual_edges = [
   ('ngc902', 'ngc4087'),
   ('ngc728', 'ngc1872'),
   ('brumerebus', 'antlejos'),
   ('kiwi', 'suna'),
   ('pellmell', 'fulcrum'),
   ('nasona', 'oriantis'),
   ('tau_ceti', 'arcanis'),
   ('ngc8338', 'unicorn'),
   #('carrza', 'tepvin'),
   #('tepvin', 'hakoi'),
]

from graphmod import ssys_jmp
from virtual_edges import add_virtual_edges

add_virtual_edges(ssys_jmp, virtual_edges)



#!/usr/bin/env python3

import argparse
import os
import sys
import logging
import json
import mimetypes
import time
import urllib.parse
import urllib.request
import urllib.error

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')

# Default MIME type for unknown files
DEFAULT_MIME_TYPE = "application/octet-stream"

def parse_args():
   parser = argparse.ArgumentParser(description="Upload a release and assets to a Forgejo (Gitea-compatible) instance.")
   parser.add_argument("--tag", required=True, help="Release tag to create/update.")
   parser.add_argument("--asset-dir", required=True, help="Directory containing assets to upload.")
   parser.add_argument("--title", help="Release title (defaults to tag).")
   parser.add_argument("--notes", help="Release notes (inline text).")
   parser.add_argument("--notes-file", help="Release notes file (overrides --notes).")
   parser.add_argument("--owner", default=os.environ.get("FORGEJO_OWNER", "naev"), help="Repository owner.")
   parser.add_argument("--repo", default=os.environ.get("FORGEJO_REPO", "naev"), help="Repository name.")
   parser.add_argument("--api-url", default=os.environ.get("FORGEJO_API_URL", "https://codeberg.org/api/v1"), help="API base URL.")
   parser.add_argument("--prerelease", action="store_true", help="Mark release as prerelease.")
   parser.add_argument("--overwrite", action="store_true", help="Delete existing release with same tag first.")
   parser.add_argument("--hide-archive-link", action="store_true", help="Hide auto-generated source archives on the release.")
   return parser.parse_args()

def get_token():
   token = os.environ.get("FORGEJO_TOKEN") or os.environ.get("CODEBERG_TOKEN") or os.environ.get("GITHUB_TOKEN")
   if not token:
      logging.error("FORGEJO_TOKEN (or CODEBERG_TOKEN / GITHUB_TOKEN) must be set")
      sys.exit(1)
   return token

def api_request(url, method="GET", headers=None, data=None):
   if headers is None:
      headers = {}

   if data is not None:
      if isinstance(data, (dict, list)):
         data = json.dumps(data).encode("utf-8")
         headers["Content-Type"] = "application/json"
      elif isinstance(data, str):
         data = data.encode("utf-8")

   req = urllib.request.Request(url, data=data, headers=headers, method=method)

   try:
      with urllib.request.urlopen(req) as response:
         if response.status == 204:
            return None
         return json.loads(response.read().decode("utf-8"))
   except urllib.error.HTTPError as e:
      # Return the error response body if possible for debugging
      try:
         err_body = e.read().decode("utf-8")
         logging.error(f"HTTP Error {e.code}: {e.reason} - {err_body}")
      except:
         logging.error(f"HTTP Error {e.code}: {e.reason}")
      raise

def retry_api_request(url, method="GET", headers=None, data=None, attempts=5):
   delay = 1
   for i in range(attempts):
      try:
         return api_request(url, method, headers, data)
      except urllib.error.HTTPError as e:
         if i == attempts - 1:
            raise
         logging.warning(f"Attempt {i+1}/{attempts} failed, retrying in {delay}s...")
         time.sleep(delay)
         delay *= 2
   return None

def main():
   args = parse_args()
   token = get_token()

   if not os.path.isdir(args.asset_dir):
      logging.error(f"Asset dir '{args.asset_dir}' not found")
      sys.exit(1)

   headers = {
      "Authorization": f"token {token}",
   }

   base_url = args.api_url.rstrip("/")
   repo_url = f"{base_url}/repos/{args.owner}/{args.repo}"

   # Prepare body
   body_content = args.notes if args.notes else ""
   if args.notes_file:
      try:
         with open(args.notes_file, "r") as f:
            body_content = f.read()
      except Exception as e:
         logging.error(f"Failed to read notes file: {e}")
         sys.exit(1)

   title = args.title if args.title else args.tag

   # Overwrite logic
   if args.overwrite:
      try:
         # Check if release exists
         try:
            release = api_request(f"{repo_url}/releases/tags/{args.tag}", headers=headers)
            if release:
               release_id = release['id']
               logging.info(f"Deleting existing release {release_id} for tag {args.tag}")
               retry_api_request(f"{repo_url}/releases/{release_id}", method="DELETE", headers=headers)
         except urllib.error.HTTPError as e:
            if e.code != 404:
               raise
      except Exception as e:
         logging.warning(f"Failed to check/delete existing release: {e}")

   # Create Release
   payload = {
      "tag_name": args.tag,
      "name": title,
      "body": body_content,
      "draft": False,
      "prerelease": args.prerelease,
      "hide_archive_links": args.hide_archive_link
   }

   try:
      logging.info(f"Creating release for tag {args.tag}")
      release = retry_api_request(f"{repo_url}/releases", method="POST", headers=headers, data=payload)
      release_id = release['id']
      logging.info(f"Created release id {release_id}")
   except Exception as e:
      logging.error(f"Failed to create release: {e}")
      sys.exit(1)

   # Upload Assets
   for filename in os.listdir(args.asset_dir):
      filepath = os.path.join(args.asset_dir, filename)
      if os.path.isfile(filepath):
         # URL encode the filename
         encoded_name = urllib.parse.quote(filename)
         logging.info(f"Uploading asset: {filename} -> {encoded_name}")

         upload_url = f"{repo_url}/releases/{release_id}/assets?name={encoded_name}"

         # Read file content
         with open(filepath, "rb") as f:
            file_data = f.read()

         # Guess MIME type or fallback
         content_type, _ = mimetypes.guess_type(filename)
         if content_type is None:
            content_type = DEFAULT_MIME_TYPE

         asset_headers = headers.copy()
         asset_headers["Content-Type"] = content_type

         try:
            retry_api_request(upload_url, method="POST", headers=asset_headers, data=file_data)
         except Exception as e:
            logging.error(f"Failed to upload asset {filename}: {e}")
            sys.exit(1)

   logging.info("Release upload complete.")

if __name__ == "__main__":
   main()



#!/usr/bin/env python3


from sys import stdout, stderr, stdin
import math

from getconst import PHYSICS_SPEED_DAMP
from arg_to_obj import arg_to_naev_obj
from slst import Slst


out = lambda x: stdout.write(x + '\n')

def line_drawing(gith = False, colour = False, term = False):
   head, rule, line, alt_line = Slst(), Slst(), Slst(), Slst()

   if colour or term:
      head += [ ('|', ' ', 1) ]
      [a.append(('', ' |', -1)) for a in [rule, line, alt_line]]
      if colour:
         colour, alt_col, defc = '\033[30;1m', '\033[34m', '\033[0m'
         rule += [ ('', colour, 1), ('', defc, -1) ]
         line += [ ('|', colour + '|' + defc) ]
         alt_line += [ ('', alt_col, 1), ('', defc, -1) ]
         #alt_line += [ ('|', (colour if term else alt_col) + '|' + defc) ]
      if term:
         rule += [
            ('| -', '|--'), ('- |',  '--|'),
            ('-|-',  "-\N{BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL}-" ),
            ('|-',   "\N{BOX DRAWINGS LIGHT DOWN AND RIGHT}-"           ),
            ('-|',   "-\N{BOX DRAWINGS LIGHT VERTICAL AND LEFT}"        ),
            ('^|',   "\N{BOX DRAWINGS LIGHT VERTICAL AND RIGHT}"        ),
            ('-',    "\N{BOX DRAWINGS LIGHT HORIZONTAL}"                ),
         ]
         line += [ ('|', "\N{BOX DRAWINGS LIGHT VERTICAL}") ]
         alt_line += [
            #('|', "\N{BOX DRAWINGS LIGHT VERTICAL AND RIGHT}", 1),
            #('|', "\N{BOX DRAWINGS LIGHT VERTICAL AND LEFT}", -1),
            #('|', "\N{BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL}")
            ('|', "\N{BOX DRAWINGS LIGHT VERTICAL}")
         ]
      lines = lambda f: alt_line if f else line
   else:
      if gith:
         rule += [ (7*'-', '-'), (3*'-', '-'), (2*'-', '-'), ('-', '---') ]
         line += [ (7*' ', ' '), (3*' ', ' '), (2*' ', ' ') ]
      lines = lambda _: line
   head += line
   if term:
      head = Slst(head[0:1]) + rule + [
         ("\N{BOX DRAWINGS LIGHT VERTICAL AND LEFT}","\N{BOX DRAWINGS LIGHT DOWN AND LEFT}"),
         ("\N{BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL}","\N{BOX DRAWINGS LIGHT DOWN AND HORIZONTAL}")
      ] + line[-1:]
   return head, rule, lines

def field( a, f ):
   try:
      res = a[f]
      if type(res) == type([]):
         res = res[0]

      if type(res) == type(()):
         res = res[0]

      if type(res) == type(''):
         res = float(res)
   except:
      res = 0.0
   return res

accel =        lambda a: field(a, 'accel')
speed =        lambda a: field(a, 'speed')
eml =          lambda a: field(a, 'engine_limit')
turn =         lambda a: field(a, 'turn')

maxspeed =     lambda a: speed(a) + accel(a) / PHYSICS_SPEED_DAMP
fullsptime =   lambda a: maxspeed(a) / accel(a) if accel(a) else 0.0
radius =       lambda a: round(maxspeed(a) / (turn(a)/180.0*math.pi))
fullspdist =   lambda a: round(maxspeed(a) * fullsptime(a) / 2.0)
turntime =     lambda a: 180.0 / turn(a)

def key( A ):
   (a, _) = A
   return (maxspeed(a), accel(a))

def fmt( f ):
   return round(f, 2)

def fmt4( n ):
   s = str(n)
   return (4-len(s))*' ' + ' ' + s + ' '

def l( s ):
   if int(s) == s:
      s = int(s)
   s = str(s)
   if '.' not in s:
      s += '.'
      m = ' '
   else:
      m = '.'
   a, b = tuple(s.split('.', 1))
   return ' | ' + (3-len(a))*' ' + a + m + b + (2-len(b))*' '

def main( args, gith = False, colour = False, term = False, autostack = False,
      combine = False, nosort = False, good_only = False ):

   L = [ (o.to_dict(), o.shortname()) for o in arg_to_naev_obj(args, combine, autostack, good_only) ]

   stderr.write('\n')
   if not nosort:
      L.sort(key = key, reverse = True)

   head, rule, lines = line_drawing(gith, colour, term)
   if colour:
      C = [ 'eml   \n(t)   ', 'drift \nspeed ', 'max   \nspeed ', 'accel \n      ',
            'fullsp\n(s)   ', 'fullsp\n(km)  ', 'turn  \n(°/s) ', 'turn  \nradius',
            '\xbdturn \n(s)   ']
      if term:
         C = [len(s.split('\n')[0])*'-' + '\n' + s for s in C]
   elif gith:
      C = ['eml (t)', 'drift speed ', 'max speed', 'accel', 'fullsp (s)',
         'fullsp (km)', 'turn (°/s)', 'turn radius', '1/2 turn (s)']
   else:
      C = [' eml  ', 'dr.sp.', 'max sp', 'accel ', 'fsp(s)',
         'fsp.km', ' turn ', 'radius', '1/2turn(s)']

   N = max([3] + [len(n) for (_, n) in L])
   C = [s.split('\n') for s in C]
   while [L for L in C if L!=[]] != []:
      lin = '| '+ N*' ' + ' | ' + ' | '.join([(L+[''])[0] for L in C])
      out(head(lin))
      C = [L[1:] for L in C]

   out(rule('| ---' + (N-3)*'-' + len(C)*' | ------'))
   for count, (k, n) in enumerate(L):
      if accel(k) != 0:
         acc = '| ' + n + (N-len(n))*' ' + ' | '
         acc += fmt4(int(eml(k))) + l(speed(k)) + l(fmt(maxspeed(k)))
         acc += l(accel(k)) + l(fmt(fullsptime(k))) + l(fmt(fullspdist(k)))
         acc += l(turn(k)) + l(radius(k)) + l(fmt(turntime(k)))
         out( lines ((count%4) == 0) (acc) )

if __name__ == '__main__':
   import argparse

   parser = argparse.ArgumentParser(
      formatter_class = argparse.RawTextHelpFormatter,
      description = 'By default, outputs text aligned markdown table comparing the engines respective values, with some new derived fields.',
      epilog = '\nTypical usage (from naev root dir) :\n > find dat/outfits/core_engine/ -name "*.xml" | ./utils/outfits/engines_stats.py'
   )
   parser.add_argument('-g', '--github', action = 'store_true', help = 'unaligned (therefore smaller) valid github md, for use in posts.')
   parser.add_argument('-c', '--colour', action = 'store_true', help = 'coloured terminal output. You can pipe to "less -RS" if the table is too wide.')
   parser.add_argument('-t', '--term', action = 'store_true', help = 'coloured terminal output with extended table characters.')
   parser.add_argument('-A', '--autostack', action = 'store_true', help = 'Outfits are presented both alone and auto-stacked.')
   parser.add_argument('-C', '--combinations', action = 'store_true', help = 'also display all the combinations.' )
   parser.add_argument('-G', '--good', action = 'store_true', help = "Like -C, but good comb. only: don't comb when pri smaller." )
   parser.add_argument('-N', '--no-sort', action = 'store_true', help = "don't sort wrt. max speed" )
   parser.add_argument('-f', '--files', action = 'store_true', help = 'read file list on stdin. Applies when no args.')
   parser.add_argument('filename', nargs = '*', help = """An outfit with ".xml" or ".mvx" extension, else will be ignored.
Can also be two outfits separated by \'+\', or an outfit prefixed with \'2x\' or suffixed with \'x2\'.""")
   args = parser.parse_args()
   if args.term and args.colour:
      print('Warning: -c subsumed by -t', file = stderr, flush = True)

   if args.github and (args.colour or args.term):
      print('Warning: Ignored conflicting: -g and -' +
         ('t' if args.term else 'c'), file = stderr, flush = True)
      args.github = args.colour = args.term = False

   if args.autostack and args.good:
      print('Warning: -A is subsumed by -G', file = stderr, flush = True)
      args.autostack = False
   elif args.autostack and args.combinations:
      print('Warning: -A is subsumed by -C', file = stderr, flush = True)
      args.autostack = False

   if args.good and args.combinations:
      print('Warning: -C is subsumed by -G', file = stderr, flush = True)

   if args.files or args.filename == []:
      args.filename += [l.strip() for l in stdin.readlines()]
   args.filename = [f for f in args.filename if f[-4:] in {".mvx", ".xml"}]
   main(args.filename, args.github, args.colour or args.term, args.term,
      args.autostack, args.combinations or args.good, args.no_sort, args.good)
else:
   raise Exception('This module is only intended to be used as main.')



#!/usr/bin/env python3

import os
from sys import stderr, exit
import re

script_dir = os.path.dirname( __file__ )
const_dir = os.path.realpath(os.path.join( script_dir, '..', '..', 'dat', 'constants.lua' ))

try:
   with open(const_dir, 'rt') as fp:
      found = re.search('PHYSICS_SPEED_DAMP *= *(.*),', fp.read()).group(1)
except:
   stderr.write('Err: '+str(const_dir)+' not found!\n')
   sys.exit(1)

try:
   PHYSICS_SPEED_DAMP = eval(found)
except:
   stderr.write(
      "Can't read value after \"PHYSICS_SPEED_DAMP = \" in " + str(const_dir) + '.\n' )
   sys.exit(1)

if __name__ == '__main__':
   stderr.write('PHYSICS_SPEED_DAMP = ' + str(PHYSICS_SPEED_DAMP) + '\n')



#!/usr/bin/env python3

import argparse
import re
import os
import sys

def parse_api( filename ):
   with open(filename, 'r', encoding='utf-8') as f:
      data = f.read()

   result = re.findall(r"static const luaL_Reg.*?\[\] = {(.+?)};", data, re.S)
   assert( len(result)==1 )
   result = re.findall( "\"(.+?)\"", result[0] )
   funclist = []
   for r in result:
      funclist += [r.replace("\"","")]

   return funclist

def generate_api( modname, filename ):
   api = parse_api( filename )
   api.sort()

   apilist = ""
   for a in api:
      apilist += f"         {a} = {{}},\n"

   return f"""stds.{modname} = {{
   read_globals = {{
      {modname} = {{
         fields = {{
{apilist}         }},
      }},
   }}
}}"""

if __name__ == "__main__":
   parser = argparse.ArgumentParser( description='Generates API for luacheckrc from a C source file.' )
   parser.add_argument('filename', metavar='FILENAME', nargs='+', type=str, help='Name of the file name to parse.')
   parser.add_argument('--output', default=None, help="Name of the output file.")
   args = parser.parse_args()

   if not args.output:
      output = sys.stdout
   else:
      output = open(args.output, 'w', encoding='utf-8')

   modlist = []
   output.write("""-- THIS FILE IS GENERATED BY utils/luacheckrc_extractor.py
-- ALL LOCAL CHANGES WILL BE LOST ON REGENERATION, SO DO NOT MODIFY PLEASE
local stds = {}
""")

   # write all the standard modules
   for filename in args.filename:
      modname = os.path.basename(filename).replace('nlua_','').replace('.c','')
      modlist += [modname]
      output.write( generate_api( modname, filename )+'\n' )

   # meta-module for naev
   for m in modlist:
      if m != 'naev':  # Do not create a recursive data structure!
         output.write(f"stds.naev.read_globals.naev.fields.{m} = stds.{m}.read_globals.{m}\n")

   output.write("return stds\n")



#!/usr/bin/env python3
import json
import os
import re
import sys

HEADER = r"""# Naev Credits
# Copyright (C) YEAR Naev Dev Team
# This file is distributed under the same licence as the naev package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: https://github.com/naev/naev/issues\n"
"POT-Creation-Date: YEAR-MO-DA HO:MI+ZONE\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@li.org>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"""

with open(sys.argv[1],"w") as fout:
   fout.write(HEADER)
   for fn in sys.argv[2:]:
      with open(fn) as f:
         fn = re.sub('.*/dat', 'dat', fn)
         for i, line in enumerate(f):
            line = line.rstrip('\r\n')
            if line and not line.startswith('['):
               quoted_escaped_line = json.dumps(line, ensure_ascii=False)
               fout.write('#: {}:{}\nmsgid {}\nmsgstr ""\n\n'.format(fn, i+1, quoted_escaped_line))
